from typing import Any, Callable, Mapping, Tuple

from ..types import TensorKind
from .base import TensorSchema
from .dense import DenseTensorSchema
from .mapped import MappedTensorSchema
from .ragged import RaggedArray, RaggedTensorSchema
from .sparse import SparseArray, SparseTensorSchema
from .sparse_to_dense import SparseToDenseTensorSchema

# mapping (is_sparse, tensor_kind): TensorSchema factory
TensorSchemaFactories: Mapping[
    Tuple[bool, TensorKind], Callable[..., TensorSchema[Any]]
] = {
    (False, TensorKind.DENSE): DenseTensorSchema,
    (True, TensorKind.DENSE): SparseToDenseTensorSchema,
    (True, TensorKind.RAGGED): RaggedTensorSchema,
    (True, TensorKind.SPARSE_COO): SparseTensorSchema,
    (True, TensorKind.SPARSE_CSR): SparseTensorSchema,
}



from pkg_resources import DistributionNotFound, get_distribution  # type: ignore

# TODO change pkg_resources with importlib.metadata, as described here
#  https://pypi.org/project/setuptools-scm/#:~:text=Retrieving%20package%20version%20at%20runtime, when we stop
#  supporting Python 3.7. We the aforementioned change, we can avoid the 100ms overhead during import of the package.

try:
    __version__ = get_distribution("tiledb-ml").version
except DistributionNotFound:
    __version__ = ""



"""Tests for TileDB integration with PyTorch Data API."""
from operator import methodcaller

import numpy as np
import pytest
import torch
import torchdata

from tiledb.ml.readers.pytorch import PyTorchTileDBDataLoader, TensorKind

from .utils import (
    ArraySpec,
    assert_tensors_almost_equal_array,
    ingest_in_tiledb,
    validate_tensor_generator,
)

non_key_dim_dtype = [np.dtype(np.int32)]
if hasattr(torch, "nested"):
    non_key_dim_dtype.append(np.dtype(np.float32))


class TestPyTorchTileDBDataLoader:
    @pytest.mark.parametrize(
        "spec", ArraySpec.combinations(non_key_dim_dtype=non_key_dim_dtype)
    )
    @pytest.mark.parametrize("batch_size", [8, None])
    @pytest.mark.parametrize("shuffle_buffer_size", [0, 16])
    @pytest.mark.parametrize("num_workers", [0, 2])
    def test_dataloader(
        self, tmpdir, spec, batch_size, shuffle_buffer_size, num_workers
    ):
        def test(*all_array_params):
            try:
                dataloader = PyTorchTileDBDataLoader(
                    *all_array_params,
                    shuffle_buffer_size=shuffle_buffer_size,
                    batch_size=batch_size,
                    num_workers=num_workers,
                )
            except NotImplementedError:
                assert num_workers and (
                    torchdata.__version__ < "0.4"
                    or spec.tensor_kind is not TensorKind.DENSE
                )
            else:
                assert isinstance(dataloader, torch.utils.data.DataLoader)
                validate_tensor_generator(
                    dataloader,
                    [params.tensor_schema for params in all_array_params],
                    [spec] * len(all_array_params),
                    batch_size,
                )
                # ensure the dataloader can be iterated again
                n1 = sum(1 for _ in dataloader)
                assert n1 != 0
                n2 = sum(1 for _ in dataloader)
                assert n1 == n2

        with ingest_in_tiledb(tmpdir, spec) as (x_params, x_data):
            # load tensors from single array
            test(x_params)
            with ingest_in_tiledb(tmpdir, spec) as (y_params, y_data):
                # load tensors from two arrays
                test(x_params, y_params)

    @pytest.mark.parametrize(
        "spec",
        ArraySpec.combinations(
            sparse=[True],
            non_key_dim_dtype=[np.dtype(np.int32)],
            tensor_kind=[TensorKind.SPARSE_CSR],
        ),
    )
    @pytest.mark.parametrize("batch_size", [8, None])
    @pytest.mark.parametrize("shuffle_buffer_size", [0, 16])
    def test_csr(self, tmpdir, spec, batch_size, shuffle_buffer_size):
        with ingest_in_tiledb(tmpdir, spec) as (params, data):
            try:
                dataloader = PyTorchTileDBDataLoader(
                    params,
                    shuffle_buffer_size=shuffle_buffer_size,
                    batch_size=batch_size,
                )
            except ValueError as ex:
                assert str(ex) == "Cannot generate CSR tensors for 3D array"
                ndim = len(spec.shape)
                assert ndim > 3 or ndim == 3 and batch_size is not None
            else:
                assert isinstance(dataloader, torch.utils.data.DataLoader)
                validate_tensor_generator(
                    dataloader, params.tensor_schema, spec, batch_size
                )
                # ensure the dataloader can be iterated again
                n1 = sum(1 for _ in dataloader)
                assert n1 != 0
                n2 = sum(1 for _ in dataloader)
                assert n1 == n2

    @pytest.mark.parametrize(
        "spec", ArraySpec.combinations(non_key_dim_dtype=non_key_dim_dtype)
    )
    def test_unequal_key_ranges(self, tmpdir, spec):
        y_spec = ArraySpec(
            # Add one extra key on Y
            shape=(spec.shape[0] + 1, *spec.shape[1:]),
            sparse=spec.sparse,
            key_dim=spec.key_dim,
            key_dim_dtype=spec.key_dim_dtype,
            non_key_dim_dtype=spec.non_key_dim_dtype,
            num_fields=spec.num_fields,
            tensor_kind=spec.tensor_kind,
        )
        with ingest_in_tiledb(tmpdir, spec) as (x_params, x_data):
            with ingest_in_tiledb(tmpdir, y_spec) as (y_params, y_data):
                with pytest.raises(ValueError) as ex:
                    PyTorchTileDBDataLoader(x_params, y_params)
                assert "All arrays must have the same key range" in str(ex.value)

    @pytest.mark.parametrize(
        "spec",
        ArraySpec.combinations(non_key_dim_dtype=non_key_dim_dtype, num_fields=[0]),
    )
    @pytest.mark.parametrize("batch_size", [8, None])
    def test_order(self, tmpdir, spec, batch_size):
        """Test we can read the data in the same order as written.

        The order is guaranteed only for sequential processing (num_workers=0) and
        no shuffling (shuffle_buffer_size=0).
        """
        to_dense = methodcaller("to_dense")
        with ingest_in_tiledb(tmpdir, spec) as (params, data):
            dataloader = PyTorchTileDBDataLoader(params, batch_size=batch_size)
            # since num_fields is 0, fields are all the array attributes of each array
            # the first item of each batch corresponds to the first attribute (="data")
            batches = [tensors[0] for tensors in dataloader]
            assert_tensors_almost_equal_array(
                batches, data, params.tensor_schema.kind, batch_size, to_dense
            )



"""Functionality for saving and loading Sklearn models as TileDB arrays"""

import pickle
from typing import Optional

import sklearn
from sklearn import config_context
from sklearn.base import BaseEstimator

import tiledb

from ._base import Meta, TileDBArtifact, Timestamp


class SklearnTileDBModel(TileDBArtifact[BaseEstimator]):
    """
    Class that implements all functionality needed to save Sklearn models as
    TileDB arrays and load Sklearn models from TileDB arrays.
    """

    Name = "SKLEARN"
    Version = sklearn.__version__

    def __init__(
        self,
        uri: str,
        namespace: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        model: Optional[BaseEstimator] = None,
    ):
        super().__init__(uri, namespace, ctx, model)

    def save(self, *, meta: Optional[Meta] = None) -> None:
        """
        Save a Sklearn model as a TileDB array.

        :param meta: Extra metadata to save in a TileDB array.
        """
        if self.artifact is None:
            raise RuntimeError("Model is not initialized")

        # Serialize model
        serialized_model = self._serialize_model()

        # Create TileDB model array
        if not tiledb.array_exists(uri=self.uri):
            self._create_array(fields=["model"])

        self._write_array(model_params={"model": serialized_model}, meta=meta)

    def load(self, *, timestamp: Optional[Timestamp] = None) -> BaseEstimator:
        """
        Load switch, i.e, decide between __load (TileDB-ML<=0.8.0) or __load_v2 (TileDB-ML>0.8.0).

        Load a Sklearn model from a TileDB array.
        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        :return: A Sklearn model object.
        """
        with tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp) as model_array:
            if self._use_legacy_schema(model_array):
                return self.__load_legacy(model_array)
            else:
                return self.__load(model_array)

    @staticmethod
    def __load_legacy(model_array: tiledb.Array) -> BaseEstimator:
        return pickle.loads(model_array[:]["model_params"].item(0))

    def __load(self, model_array: tiledb.Array) -> BaseEstimator:
        return self._get_model_param(model_array, "model")

    def preview(self, *, display: str = "text") -> str:
        """
        Create a text representation of the model.

        :param display. If ‘diagram’, estimators will be displayed as a diagram in an
            HTML format when shown in a jupyter notebook. If ‘text’, estimators will be
            displayed as text.
        :return. A string representation of the models internal configuration.
        """
        if self.artifact:
            with config_context(display=display):
                return str(self.artifact)
        else:
            return ""

    def _serialize_model(self) -> bytes:
        """
        Serialize a Sklearn model with pickle.

        :return: Pickled Sklearn model.
        """
        return pickle.dumps(self.artifact, protocol=4)



"""Base Class for saving and loading machine learning models."""

import glob
import os
import pickle
import platform
import time
from abc import ABC, abstractmethod
from typing import (
    Any,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import numpy as np

import tiledb

from .. import __version__
from ._cloud_utils import get_cloud_uri, update_file_properties
from ._file_properties import ModelFileProperties

Artifact = TypeVar("Artifact")
Meta = Mapping[str, Any]
Timestamp = Tuple[int, int]

Weights = Union[Sequence[np.ndarray], Mapping[str, Any]]


class TileDBArtifact(ABC, Generic[Artifact]):
    """
    This is the base class for all TileDB model storage functionalities, i.e,
    store machine learning models (Tensorflow, PyTorch, etc) as TileDB arrays.
    """

    Name: str
    Version: str

    def __init__(
        self,
        uri: str,
        namespace: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        artifact: Optional[Artifact] = None,
    ):
        """
        Base class for saving machine learning models as TileDB arrays
        and loading machine learning models from TileDB arrays. In case we need to interact
        with TileDB-Cloud we have to pass user's TileDB-Cloud namespace. If we don't
        models will be saved locally.

        :param uri: TileDB array uri
        :param namespace: In case we want to interact (save, load, update, check) with
            models on TileDB-Cloud we need the user's namespace on TileDB-Cloud.
            Moreover, array's uri must have an s3 prefix.
        :param ctx: TileDB Context.
        :param artifact: Machine learning artifact (e.g. machine learning model) based on the framework we are using.
        """
        self.namespace = namespace
        self.ctx = ctx
        self.artifact = artifact
        self.uri = get_cloud_uri(uri, namespace) if namespace else uri
        self._file_properties = {
            ModelFileProperties.TILEDB_ML_MODEL_ML_FRAMEWORK.value: self.Name,
            ModelFileProperties.TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION.value: self.Version,
            ModelFileProperties.TILEDB_ML_MODEL_STAGE.value: "STAGING",
            ModelFileProperties.TILEDB_ML_MODEL_PYTHON_VERSION.value: platform.python_version(),
            ModelFileProperties.TILEDB_ML_MODEL_PREVIEW.value: self.preview(),
            ModelFileProperties.TILEDB_ML_MODEL_VERSION.value: __version__,
        }

    @abstractmethod
    def save(self, *, meta: Optional[Meta] = None) -> None:
        """Abstract method that saves a machine learning model as a TileDB array.
        :param meta: Extra metadata to save in a TileDB array.
        """

    @abstractmethod
    def load(self, *, timestamp: Optional[Timestamp] = None) -> Artifact:
        """Abstract method that loads a machine learning model from a TileDB array.

        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        """

    def get_weights(self, timestamp: Optional[Timestamp] = None) -> Weights:
        """
        Returns model's weights. Works for Tensorflow Keras and PyTorch
        """
        with tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp) as model_array:
            return cast(Weights, self._get_model_param(model_array, "model"))

    def get_optimizer_weights(self, timestamp: Optional[Timestamp] = None) -> Weights:
        """
        Returns optimizer's weights. Works for Tensorflow Keras and PyTorch
        """
        with tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp) as model_array:
            return cast(Weights, self._get_model_param(model_array, "optimizer"))

    @abstractmethod
    def preview(self) -> str:
        """
        Creates a string representation of a machine learning model.
        """

    def _create_array(self, fields: Sequence[str]) -> None:
        """Internal method that creates a TileDB array based on the model's spec."""

        # The array will be be 1 dimensional with domain of 0 to max uint64. We use a tile extent of 1024 bytes
        dom = tiledb.Domain(
            tiledb.Dim(
                name="offset",
                domain=(0, np.iinfo(np.uint64).max - 1025),
                tile=1024,
                dtype=np.uint64,
                ctx=self.ctx,
            ),
        )

        attrs = [
            tiledb.Attr(
                name=field,
                dtype=np.uint8,
                filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                ctx=self.ctx,
            )
            for field in fields
        ]

        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=False,
            attrs=attrs,
            ctx=self.ctx,
        )

        tiledb.Array.create(self.uri, schema, ctx=self.ctx)

        # In case we are on TileDB-Cloud we have to update model array's file properties
        if self.namespace:
            update_file_properties(self.uri, self._file_properties)

    def _write_array(
        self,
        model_params: Mapping[str, bytes],
        tensorboard_log_dir: Optional[str] = None,
        meta: Optional[Meta] = None,
    ) -> None:
        if tensorboard_log_dir:
            tensorboard = self._serialize_tensorboard(tensorboard_log_dir)
        else:
            tensorboard = b""
        model_params = dict(tensorboard=tensorboard, **model_params)

        if meta is None:
            meta = {}
        if not meta.keys().isdisjoint(self._file_properties.keys()):
            raise ValueError(
                "Please avoid using file property key names as metadata keys!"
            )

        timestamp = round(time.time() * 1000)
        with tiledb.open(
            self.uri, "w", timestamp=timestamp, ctx=self.ctx
        ) as model_array:
            one_d_buffers = {}
            max_len = 0
            for key, value in model_params.items():
                one_d_buffer = np.frombuffer(value, dtype=np.uint8)
                one_d_buffer_len = len(one_d_buffer)
                one_d_buffers[key] = one_d_buffer
                # Write size only in case is greater than 0.
                if one_d_buffer_len:
                    model_array.meta[key + "_size"] = one_d_buffer_len
                if one_d_buffer_len > max_len:
                    max_len = one_d_buffer_len

            model_array[0:max_len] = {
                key: np.pad(value, (0, max_len - len(value)))
                for key, value in one_d_buffers.items()
            }
            for mapping in meta, self._file_properties:
                for key, value in mapping.items():
                    model_array.meta[key] = value

    def _get_model_param(self, model_array: tiledb.Array, key: str) -> Any:
        size_key = key + "_size"
        try:
            size = model_array.meta[size_key]
        except KeyError:
            raise Exception(
                f"{size_key} metadata entry not present in {self.uri}"
                f" (existing keys: {set(model_array.meta.keys())})"
            )
        return pickle.loads(model_array.query(attrs=(key,))[0:size][key].tobytes())

    @staticmethod
    def _serialize_tensorboard(log_dir: str) -> bytes:
        """Serialize all Tensorboard files."""
        if not os.path.exists(log_dir):
            raise ValueError(f"{log_dir} does not exist")
        tensorboard_files = {}
        for path in glob.glob(f"{log_dir}/*tfevents*"):
            with open(path, "rb") as f:
                tensorboard_files[path] = f.read()
        return pickle.dumps(tensorboard_files, protocol=4)

    def _load_tensorboard(self, model_array: tiledb.Array) -> None:
        """
        Write Tensorboard files to directory. Works for Tensorflow-Keras and PyTorch.
        """
        tensorboard_files = self._get_model_param(model_array, "tensorboard")
        for path, file_bytes in tensorboard_files.items():
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "wb") as f:
                f.write(file_bytes)

    @staticmethod
    def _use_legacy_schema(model_array: tiledb.Array) -> bool:
        # TODO: Decide based on tiledb-ml version and not on schema characteristics, like "offset".
        return str(model_array.schema.domain.dim(0).name) != "offset"



from operator import itemgetter
from typing import Any, Iterable, Sequence, Union, cast

import numpy as np
import scipy.sparse
import sparse

from .base_sparse import BaseSparseTensorSchema
from .ranges import InclusiveRange

SparseArray = Union[scipy.sparse.csr_matrix, sparse.COO]


class SparseTensorSchema(BaseSparseTensorSchema[SparseArray]):
    """
    TensorSchema for reading sparse TileDB arrays as SparseArray instances.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._query_kwargs["dims"] = self._all_dims

    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[Any, int]]
    ) -> Union[Iterable[SparseArray], Iterable[Sequence[SparseArray]]]:
        shape = list(cast(Sequence[int], self.shape))
        SparseArrayFactory = csr_matrix if len(shape) == 2 else sparse.COO
        query = self.query
        get_data = itemgetter(*self._fields)
        single_field = len(self._fields) == 1
        key_dim, *non_key_dims = self._all_dims
        non_key_dim_starts = tuple(map(itemgetter(0), self._ned[1:]))
        for key_range in key_ranges:
            # Set the shape of the key dimension equal to the current key range length
            shape[0] = len(key_range)
            field_arrays = query[key_range.min : key_range.max]
            data = get_data(field_arrays)

            # Convert coordinates from the original domain to zero-based
            # For the key (i.e. first) dimension get the indices of the keys
            coords = [key_range.indices(field_arrays.pop(key_dim))]
            # For every non-key dimension, subtract the minimum value of the dimension
            # TODO: update this for non-integer non-key dimensions
            coords.extend(
                field_arrays.pop(dim) - dim_start
                for dim, dim_start in zip(non_key_dims, non_key_dim_starts)
            )
            coords = np.array(coords)

            # yield either a single SparseArray or one SparseArray per field
            if single_field:
                yield SparseArrayFactory(coords, data, shape)
            else:
                yield tuple(SparseArrayFactory(coords, d, shape) for d in data)


def csr_matrix(
    coords: np.ndarray, data: np.ndarray, shape: Sequence[int]
) -> scipy.sparse.csr_matrix:
    return scipy.sparse.csr_matrix((data, coords), shape)



import itertools as it
import string

import numpy as np
import pytest

import tiledb
from tiledb.ml.readers.types import ArrayParams


@pytest.fixture(scope="module")
def dense_uri(tmp_path_factory):
    uri = str(tmp_path_factory.mktemp("dense"))
    schema = tiledb.ArraySchema(
        sparse=False,
        domain=tiledb.Domain(
            tiledb.Dim(name="d0", domain=(0, 9999), dtype=np.int32, tile=123),
            tiledb.Dim(name="d1", domain=(-20, 19), dtype=np.int32, tile=4),
        ),
        attrs=[
            tiledb.Attr(name="af8", dtype=np.float64),
            tiledb.Attr(name="af4", dtype=np.float32),
            tiledb.Attr(name="au1", dtype=np.uint8),
        ],
    )
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as a:
        a[:] = {
            "af8": np.random.rand(*schema.shape),
            "af4": np.random.rand(*schema.shape).astype(np.float32),
            "au1": np.random.randint(128, size=schema.shape, dtype=np.uint8),
        }
    return uri


@pytest.fixture(scope="module")
def sparse_uri(tmp_path_factory, non_zero_per_row=3):
    uri = str(tmp_path_factory.mktemp("sparse"))
    domains = [
        (1, 1000),
        (-250, 249),
        (0, 1),
        (np.datetime64("2020-01-01"), np.datetime64("2022-01-01")),
        ("a", "z"),
    ]
    schema = tiledb.ArraySchema(
        sparse=True,
        allows_duplicates=True,
        domain=tiledb.Domain(
            tiledb.Dim(name="d0", domain=domains[0], dtype=np.uint32),
            tiledb.Dim(name="d1", domain=domains[1], dtype=np.int16),
            tiledb.Dim(name="d2", domain=domains[2], dtype=np.float64),
            tiledb.Dim(name="d3", domain=domains[3], dtype=np.dtype("datetime64[D]")),
            tiledb.Dim(name="d4", domain=domains[4], dtype="ascii"),
        ),
        attrs=[
            tiledb.Attr(name="af8", dtype=np.float64),
            tiledb.Attr(name="af4", dtype=np.float32),
            tiledb.Attr(name="au1", dtype=np.uint8),
        ],
    )
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as a:
        d0 = np.concatenate(
            [np.arange(domains[0][0], domains[0][1] + 1)] * non_zero_per_row
        )
        num_cells = len(d0)
        d1 = np.random.randint(domains[1][0], domains[1][1] + 1, num_cells)
        d2 = np.random.uniform(domains[2][0], domains[2][1], num_cells)
        d3_days = (domains[3][1] - domains[3][0]).astype(int)
        d3 = domains[3][0] + np.random.randint(0, d3_days, num_cells)
        # TODO: increase 8 to something larger after sh-19349 is fixed
        d4 = np.random.choice([c * 8 for c in string.ascii_lowercase], num_cells)
        a[d0, d1, d2, d3, d4] = {
            "af8": np.random.rand(num_cells),
            "af4": np.random.rand(num_cells).astype(np.float32),
            "au1": np.random.randint(128, size=num_cells, dtype=np.uint8),
        }
    return uri


def parametrize_fields(*fields, num=3):
    return pytest.mark.parametrize(
        "fields",
        list(it.chain.from_iterable(it.combinations(fields, i) for i in range(num))),
    )


@pytest.mark.parametrize(
    "key_dim,memory_budget,dim_selectors",
    [
        ("d0", 160_000, {}),
        ("d0", 160_000, {"d0": slice(1000, 9000)}),
        ("d0", 160_000, {"d0": slice(None, 9000)}),
        ("d0", 160_000, {"d0": slice(1000, None)}),
        ("d0", 160_000, {"d1": slice(-10, 10)}),
        ("d0", 160_000, {"d1": slice(-10, None)}),
        ("d0", 160_000, {"d1": slice(None, 10)}),
        ("d0", 160_000, {"d1": list(range(-10, 10, 3))}),
        ("d0", 160_000, {"d0": slice(1000, 9000), "d1": slice(-10, 10)}),
        ("d0", 160_000, {"d0": slice(None, 9000), "d1": slice(-10, None)}),
        ("d0", 160_000, {"d0": slice(1000, None), "d1": slice(None, 10)}),
        ("d1", 700_000, {}),
        ("d1", 700_000, {"d1": slice(-10, 10)}),
        ("d1", 700_000, {"d1": slice(None, 10)}),
        ("d1", 700_000, {"d1": slice(-10, None)}),
        ("d1", 700_000, {"d0": slice(1000, 9000)}),
        ("d1", 700_000, {"d0": slice(None, 9000)}),
        ("d1", 700_000, {"d0": slice(1000, None)}),
        ("d1", 700_000, {"d0": list(range(100, 5000, 3))}),
        ("d1", 700_000, {"d1": slice(-10, 10), "d0": slice(1000, 9000)}),
        ("d1", 700_000, {"d1": slice(None, 10), "d0": slice(None, 9000)}),
        ("d1", 700_000, {"d1": slice(-10, None), "d0": slice(1000, None)}),
    ],
)
@parametrize_fields("d0", "d1", "af8", "af4", "au1")
def test_max_partition_weight_dense(
    dense_uri, fields, key_dim, memory_budget, dim_selectors
):
    config = {"py.max_incomplete_retries": 0, "sm.memory_budget": memory_budget}
    with tiledb.open(dense_uri, config=config) as array:
        _test_max_partition_weight(array, fields, key_dim, dim_selectors)


@pytest.mark.parametrize(
    "key_dim,memory_budget,dim_selectors",
    [
        ("d0", 2048, {}),
        ("d0", 4096, {}),
        ("d1", 2048, {}),
        ("d1", 4096, {}),
        ("d3", 2048, {}),
        ("d3", 4096, {}),
        ("d4", 2048, {}),
        ("d4", 4096, {}),
        ("d0", 2048, {"d1": list(range(-100, 100, 2))}),
        ("d0", 3072, {"d0": slice(100, 900)}),
        ("d0", 4096, {"d1": slice(None, 0), "d2": slice(0.5, None)}),
        ("d1", 2048, {"d3": slice(None, np.datetime64("2021-12-31"))}),
        ("d1", 3072, {"d1": slice(-200, None)}),
        ("d1", 4096, {"d2": slice(0.25, 0.75), "d4": slice(b"q", None)}),
        ("d3", 2048, {"d1": slice(100, None)}),
        ("d3", 3072, {"d3": slice(np.datetime64("2021-06-01"), None)}),
        ("d3", 4096, {"d0": list(range(1, 1000, 5)), "d4": slice(b"a", b"k")}),
        ("d4", 2048, {"d1": list(range(0, 200, 4)), "d2": slice(None, 0.25)}),
        ("d4", 3072, {"d4": slice(None, b"n")}),
        ("d4", 4096, {"d3": slice(np.datetime64("2021-01-01"), None)}),
    ],
)
@parametrize_fields("d0", "d1", "d2", "d3", "d4", "af8", "af4", "au1")
def test_max_partition_weight_sparse(
    sparse_uri, fields, key_dim, memory_budget, dim_selectors
):
    config = {
        "py.max_incomplete_retries": 0,
        "py.init_buffer_bytes": memory_budget,
    }
    with tiledb.open(sparse_uri, config=config) as array:
        _test_max_partition_weight(array, fields, key_dim, dim_selectors)


def _test_max_partition_weight(array, fields, key_dim, dim_selectors):
    schema = ArrayParams(array, key_dim, fields, dim_selectors).tensor_schema
    max_weight = schema.max_partition_weight
    key_ranges = list(schema.key_range.partition_by_weight(max_weight))
    for i, key_range in enumerate(key_ranges):
        # query succeeds without incomplete retries
        schema.query[key_range.min : key_range.max]

        if i < len(key_ranges) - 1:
            # querying a larger slice should fail
            with pytest.raises(tiledb.TileDBError) as ex:
                schema.query[key_range.min : key_ranges[i + 1].min]
            assert "py.max_incomplete_retries" in str(ex.value)



"""Tests for TileDB Tensorflow Keras model save and load."""

import glob
import io
import os
import pickle
import platform
import shutil

import numpy as np
import pytest
import tensorflow as tf

import tiledb
from tiledb.ml import __version__ as tiledb_ml_version
from tiledb.ml.models.tensorflow_keras import TensorflowKerasTileDBModel

try:
    from keras.testing_infra.test_utils import (
        get_small_functional_mlp,
        get_small_sequential_mlp,
    )
except ImportError:
    from keras.testing_utils import get_small_functional_mlp, get_small_sequential_mlp

# Suppress all Tensorflow messages
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

batch_get_value = tf.keras.backend.batch_get_value


def read_files(dirpath):
    files = {}
    for path in glob.glob(f"{dirpath}/*"):
        with open(path, "rb") as f:
            files[path] = f.read()
    return files


class ConfigSubclassModel(tf.keras.Model):
    def __init__(self, hidden_units):
        super(ConfigSubclassModel, self).__init__()
        self.hidden_units = hidden_units
        self.dense_layers = [tf.keras.layers.Dense(u) for u in hidden_units]

    def call(self, inputs):
        x = inputs
        for layer in self.dense_layers:
            x = layer(x)
        return x

    def get_config(self):
        return {"hidden_units": self.hidden_units}

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def test_load_tiledb_error_with_wrong_uri():
    tiledb_model_obj = TensorflowKerasTileDBModel(uri="dummy_uri")
    with pytest.raises(tiledb.TileDBError):
        tiledb_model_obj.load(compile_model=False)


loss_optimizer_metrics = pytest.mark.parametrize(
    "loss,optimizer,metrics",
    [
        (None, None, None),
        ("binary_crossentropy", "rmsprop", "accuracy"),
        ("mean_squared_error", "rmsprop", "categorical_accuracy"),
        ("sparse_categorical_crossentropy", "sgd", "sparse_categorical_crossentropy"),
    ],
)

api = pytest.mark.parametrize(
    "api",
    [get_small_sequential_mlp, get_small_functional_mlp],
)


class TestTensorflowKerasModel:
    @loss_optimizer_metrics
    def test_exception_on_subclassed_models(self, tmpdir, loss, optimizer, metrics):
        model = ConfigSubclassModel(hidden_units=[16, 16, 10])
        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        if not model.built:
            model.build(tuple(np.random.randint(20, size=2)))
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        with pytest.raises(RuntimeError):
            tiledb_model_obj.save(include_optimizer=True if optimizer else False)

    @api
    @loss_optimizer_metrics
    def test_save_model_to_tiledb_array(self, tmpdir, api, loss, optimizer, metrics):
        model = api(num_hidden=1, num_classes=2, input_dim=3)
        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        if not model.built:
            model.build(tuple(np.random.randint(20, size=2)))
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True if optimizer else False)
        assert tiledb.array_exists(tiledb_uri)

        with pytest.raises(RuntimeError) as ex:
            tiledb_obj = TensorflowKerasTileDBModel(uri="")
            tiledb_obj.save()

        assert "Model is not initialized" in str(ex.value)

    @api
    @loss_optimizer_metrics
    def test_save_model_to_tiledb_array_predictions(
        self, tmpdir, api, loss, optimizer, metrics
    ):
        model = api(num_hidden=1, num_classes=2, input_dim=3)
        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        input_shape = tuple(np.random.randint(20, size=2))
        if not model.built:
            model.build(input_shape)
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)

        tiledb_model_obj.save(include_optimizer=True if optimizer else False)

        loaded_model = tiledb_model_obj.load(compile_model=False)

        data = np.random.rand(100, 3)

        # Assert model predictions are equal
        np.testing.assert_array_equal(loaded_model.predict(data), model.predict(data))

    @api
    @loss_optimizer_metrics
    def test_save_model_to_tiledb_array_weights(
        self, tmpdir, api, loss, optimizer, metrics
    ):
        model = api(num_hidden=1, num_classes=2, input_dim=3)
        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        input_shape = tuple(np.random.randint(20, size=2))
        if not model.built:
            model.build(input_shape)
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)

        tiledb_model_obj.save(include_optimizer=True if optimizer else False)
        loaded_model = tiledb_model_obj.load(compile_model=True if optimizer else False)

        data = np.random.rand(100, 3)

        if optimizer:
            model_opt_weights = batch_get_value(model.optimizer.weights)
            loaded_opt_weights = batch_get_value(loaded_model.optimizer.weights)

            # Assert optimizer weights are equal
            for weight_model, weight_loaded_model in zip(
                model_opt_weights, loaded_opt_weights
            ):
                np.testing.assert_array_equal(weight_model, weight_loaded_model)

            # Assert model predictions are equal
            np.testing.assert_array_equal(
                loaded_model.predict(data), model.predict(data)
            )

    @loss_optimizer_metrics
    def test_save_load_with_dense_features(self, tmpdir, loss, optimizer, metrics):
        if optimizer is None:
            pytest.skip()
        cols = [
            tf.feature_column.numeric_column("a"),
            tf.feature_column.indicator_column(
                tf.feature_column.categorical_column_with_vocabulary_list(
                    "b", ["one", "two"]
                )
            ),
        ]
        input_layers = {
            "a": tf.keras.layers.Input(shape=(1,), name="a"),
            "b": tf.keras.layers.Input(shape=(1,), name="b", dtype="string"),
        }

        fc_layer = tf.keras.layers.DenseFeatures(cols)(input_layers)
        output = tf.keras.layers.Dense(10)(fc_layer)

        model = tf.keras.Model(input_layers, output)

        model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        model_opt_weights = batch_get_value(model.optimizer.weights)
        loaded_opt_weights = batch_get_value(loaded_model.optimizer.weights)

        # Assert optimizer weights are equal
        for weight_model, weight_loaded_model in zip(
            model_opt_weights, loaded_opt_weights
        ):
            np.testing.assert_array_equal(weight_model, weight_loaded_model)

        inputs_a = np.arange(10).reshape(10, 1)
        inputs_b = np.arange(10).reshape(10, 1).astype("str")

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict({"a": inputs_a, "b": inputs_b}),
            model.predict({"a": inputs_a, "b": inputs_b}),
        )

    @loss_optimizer_metrics
    def test_save_load_with_sequence_features(self, tmpdir, loss, optimizer, metrics):
        if optimizer is None:
            pytest.skip()

        cols = [
            tf.feature_column.sequence_numeric_column("a"),
            tf.feature_column.indicator_column(
                tf.feature_column.sequence_categorical_column_with_vocabulary_list(
                    "b", ["one", "two"]
                )
            ),
        ]
        input_layers = {
            "a": tf.keras.layers.Input(shape=(None, 1), sparse=True, name="a"),
            "b": tf.keras.layers.Input(
                shape=(None, 1), sparse=True, name="b", dtype="string"
            ),
        }

        fc_layer, _ = tf.keras.experimental.SequenceFeatures(cols)(input_layers)
        x = tf.keras.layers.GRU(32)(fc_layer)
        output = tf.keras.layers.Dense(10)(x)

        model = tf.keras.Model(input_layers, output)

        model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        model_opt_weights = batch_get_value(model.optimizer.weights)
        loaded_opt_weights = batch_get_value(loaded_model.optimizer.weights)

        # Assert optimizer weights are equal
        for weight_model, weight_loaded_model in zip(
            model_opt_weights, loaded_opt_weights
        ):
            np.testing.assert_array_equal(weight_model, weight_loaded_model)

        batch_size = 10
        timesteps = 1

        values_a = np.arange(10, dtype=np.float32)
        indices_a = np.zeros((10, 3), dtype=np.int64)
        indices_a[:, 0] = np.arange(10)
        inputs_a = tf.SparseTensor(indices_a, values_a, (batch_size, timesteps, 1))

        values_b = np.zeros(10, dtype=np.str)
        indices_b = np.zeros((10, 3), dtype=np.int64)
        indices_b[:, 0] = np.arange(10)
        inputs_b = tf.SparseTensor(indices_b, values_b, (batch_size, timesteps, 1))

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict({"a": inputs_a, "b": inputs_b}, steps=1),
            model.predict({"a": inputs_a, "b": inputs_b}, steps=1),
        )

    def test_functional_model_save_load_with_custom_loss_and_metric(self, tmpdir):
        inputs = tf.keras.layers.Input(shape=(4,))
        x = tf.keras.layers.Dense(8, activation="relu")(inputs)
        outputs = tf.keras.layers.Dense(3, activation="softmax")(x)
        model = tf.keras.Model(inputs=inputs, outputs=outputs)
        custom_loss = tf.keras.layers.Lambda(lambda x: tf.keras.backend.sum(x * x))(x)
        model.add_loss(custom_loss)
        model.add_metric(custom_loss, aggregation="mean", name="custom_loss")

        model.compile(
            loss="sparse_categorical_crossentropy",
            optimizer="sgd",
            metrics=["sparse_categorical_crossentropy"],
        )

        data_x = np.random.normal(size=(32, 4))
        data_y = np.random.randint(0, 3, size=32)
        model.train_on_batch(data_x, data_y)

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        # Assert all evaluation results are the same.
        assert all(
            [
                a == pytest.approx(b, 1e-9)
                for a, b in zip(
                    model.evaluate(data_x, data_y),
                    loaded_model.evaluate(data_x, data_y),
                )
            ]
        )

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict(data_x), model.predict(data_x)
        )

    def test_save_load_for_rnn_layers(self, tmpdir):
        inputs = tf.keras.layers.Input([10, 10], name="train_input")
        rnn_layers = [
            tf.keras.layers.LSTMCell(size, recurrent_dropout=0, name="rnn_cell%d" % i)
            for i, size in enumerate([32, 32])
        ]
        rnn_output = tf.keras.layers.RNN(
            rnn_layers, return_sequences=True, name="rnn_layer"
        )(inputs)
        pred_feat = tf.keras.layers.Dense(10, name="prediction_features")(rnn_output)
        pred = tf.keras.layers.Softmax()(pred_feat)
        model = tf.keras.Model(inputs=[inputs], outputs=[pred, pred_feat])

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=False)
        loaded_model = tiledb_model_obj.load(compile_model=False)

        data = np.random.rand(50, 10, 10)

        # Assert model predictions are equal
        np.testing.assert_array_equal(loaded_model.predict(data), model.predict(data))

    def test_sequential_model_save_load_without_input_shape(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Dense(2))
        model.add(tf.keras.layers.RepeatVector(3))
        model.add(tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(3)))
        model.compile(
            loss="mean_squared_error",
            optimizer="rmsprop",
            metrics="categorical_accuracy",
            weighted_metrics="categorical_accuracy",
            sample_weight_mode="temporal",
        )
        data_x = np.random.random((1, 3))
        data_y = np.random.random((1, 3, 3))
        model.train_on_batch(data_x, data_y)

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict(data_x), model.predict(data_x)
        )

    @api
    @loss_optimizer_metrics
    def test_preview(self, tmpdir, api, loss, optimizer, metrics):
        model = api(num_hidden=1, num_classes=2, input_dim=3)
        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        # With model given as argument
        if model.built:
            tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
            s = io.StringIO()
            model.summary(print_fn=lambda x: s.write(x + "\n"))
            model_summary = s.getvalue()
            assert tiledb_model_obj.preview() == model_summary
        else:
            # Model should be built before preview it
            with pytest.raises(ValueError):
                tiledb_model_obj = TensorflowKerasTileDBModel(
                    uri=tiledb_uri, model=model
                )
                tiledb_model_obj.preview()

        # When model is None then preview returns empty string
        tiledb_model_obj_none = TensorflowKerasTileDBModel(uri=tiledb_uri, model=None)
        assert tiledb_model_obj_none.preview() == ""


class TestTensorflowKerasModelCloud:
    def test_file_properties(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))

        # Get model summary in a string
        s = io.StringIO()
        model.summary(print_fn=lambda x: s.write(x + "\n"))
        model_summary = s.getvalue()

        uri = os.path.join(tmpdir, "model_array")
        tiledb_obj = TensorflowKerasTileDBModel(uri=uri, model=model)

        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK"]
            == "TENSORFLOW KERAS"
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_STAGE"] == "STAGING"
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_PYTHON_VERSION"]
            == platform.python_version()
        )
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"]
            == tf.__version__
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_PREVIEW"] == model_summary
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_VERSION"] == tiledb_ml_version
        )

    def test_get_cloud_uri_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        uri = os.path.join(tmpdir, "model_array")

        mock_get_cloud_uri = mocker.patch(
            "tiledb.ml.models._base.get_cloud_uri", return_value=uri
        )

        _ = TensorflowKerasTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_cloud_uri.assert_called_once_with(uri, "test_namespace")

    def test_get_s3_prefix_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        uri = os.path.join(tmpdir, "model_array")

        mock_get_s3_prefix = mocker.patch(
            "tiledb.ml.models._cloud_utils.get_s3_prefix", return_value="s3 prefix"
        )

        _ = TensorflowKerasTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_s3_prefix.assert_called_once_with("test_namespace")

    def test_update_file_properties_call(self, tmpdir, mocker):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))

        # Get model summary in a string
        s = io.StringIO()
        model.summary(print_fn=lambda x: s.write(x + "\n"))
        model_summary = s.getvalue()

        uri = os.path.join(tmpdir, "model_array")

        mocker.patch("tiledb.ml.models._base.get_cloud_uri", return_value=uri)

        tiledb_obj = TensorflowKerasTileDBModel(
            uri=uri, namespace="test_namespace", model=model
        )

        mock_update_file_properties = mocker.patch(
            "tiledb.ml.models._base.update_file_properties",
            return_value=None,
        )
        mocker.patch(
            "tiledb.ml.models.tensorflow_keras.TensorflowKerasTileDBModel._write_array"
        )

        tiledb_obj.save()

        file_properties_dict = {
            "TILEDB_ML_MODEL_ML_FRAMEWORK": "TENSORFLOW KERAS",
            "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION": tf.__version__,
            "TILEDB_ML_MODEL_STAGE": "STAGING",
            "TILEDB_ML_MODEL_PYTHON_VERSION": platform.python_version(),
            "TILEDB_ML_MODEL_PREVIEW": model_summary,
            "TILEDB_ML_MODEL_VERSION": tiledb_ml_version,
        }

        mock_update_file_properties.assert_called_once_with(uri, file_properties_dict)

    def test_exception_raise_file_property_in_meta_error(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = TensorflowKerasTileDBModel(uri=tiledb_array, model=model)
        with pytest.raises(ValueError) as ex:
            tiledb_obj.save(
                meta={"TILEDB_ML_MODEL_ML_FRAMEWORK": "TILEDB_ML_MODEL_ML_FRAMEWORK"},
            )

        assert "Please avoid using file property key names as metadata keys!" in str(
            ex.value
        )

    def test_tensorboard_callback(self, mocker, tmpdir):
        mocker.patch(
            "tiledb.ml.models.tensorflow_keras.TensorflowKerasTileDBModel._use_legacy_schema",
            return_value=False,
        )

        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = TensorflowKerasTileDBModel(uri=tiledb_array, model=model)

        cb = [tf.keras.callbacks.TensorBoard(log_dir=tmpdir)]

        os.mkdir(os.path.join(tmpdir, "train"))
        with open(os.path.join(tmpdir, "train", "foo_tfevents_1"), "wb") as f:
            f.write(b"test_bytes_1")
        with open(os.path.join(tmpdir, "train", "bar_tfevents_2"), "wb") as f:
            f.write(b"test_bytes_2")

        log_files = read_files(os.path.join(tmpdir, "train"))
        assert log_files

        tiledb_obj.save(callbacks=cb)
        with tiledb.open(tiledb_array) as A:
            tb_size = A.meta["tensorboard_size"]
            assert pickle.loads(A[0:tb_size]["tensorboard"]) == {
                os.path.join(tmpdir, "train", "foo_tfevents_1"): b"test_bytes_1",
                os.path.join(tmpdir, "train", "bar_tfevents_2"): b"test_bytes_2",
            }
        shutil.rmtree(os.path.join(tmpdir, "train"))

        tiledb_obj.load(compile_model=False, callback=True)
        with open(os.path.join(tmpdir, "train", "foo_tfevents_1"), "rb") as f:
            assert f.read() == b"test_bytes_1"
        with open(os.path.join(tmpdir, "train", "bar_tfevents_2"), "rb") as f:
            assert f.read() == b"test_bytes_2"

        shutil.rmtree(os.path.join(tmpdir, "train"))



import os

import tiledb.cloud

# Your TileDB username and password, exported as environmental variables
TILEDB_USER_NAME = os.environ.get("TILEDB_USER_NAME")
TILEDB_PASSWD = os.environ.get("TILEDB_PASSWD")

# Your TileDB namespace
TILEDB_NAMESPACE = "your_tiledb_namespace"

# Your S3 bucket
S3_BUCKET = "your_s3_bucket"

IMAGES_URI = "tiledb://{}/s3://{}/mnist_images".format(TILEDB_NAMESPACE, S3_BUCKET)
LABELS_URI = "tiledb://{}/s3://{}/mnist_labels".format(TILEDB_NAMESPACE, S3_BUCKET)
MODEL_URI = "tiledb://{}/s3://{}/mnist_model".format(TILEDB_NAMESPACE, S3_BUCKET)

# The size of each slice from a image and label TileDB arrays.
IO_BATCH_SIZE = 20000


def train() -> None:
    from typing import Any, Tuple

    import numpy as np
    import torch
    import torch.nn as nn
    import torch.optim as optim

    from tiledb.ml.models.pytorch import PyTorchTileDBModel
    from tiledb.ml.readers.pytorch import PyTorchTileDBDataLoader

    class Net(nn.Module):
        def __init__(self, shape: Tuple[int, int]):
            super().__init__()
            self.flatten = nn.Flatten()
            self.linear_relu_stack = nn.Sequential(
                nn.Linear(np.product(shape), 32),
                nn.ReLU(),
                nn.Linear(32, 32),
                nn.ReLU(),
                nn.Linear(32, 10),
                nn.ReLU(),
            )

        def forward(self, x: torch.Tensor) -> Any:
            x = self.flatten(x)
            logits = self.linear_relu_stack(x)
            return logits

    with tiledb.open(IMAGES_URI) as x, tiledb.open(LABELS_URI) as y:
        train_loader = PyTorchTileDBDataLoader(x, y, batch_size=IO_BATCH_SIZE)

        net = Net(shape=(28, 28))
        criterion = nn.CrossEntropyLoss()
        optimizer = optim.SGD(net.parameters(), lr=0.01, momentum=0.5)

        training_batch_size = 200

        for epoch in range(1, 3):
            net.train()
            training_batch_idx = 0
            for batch_idx, (inputs, labels) in enumerate(train_loader):
                for train_batch_start_idx in range(
                    0, IO_BATCH_SIZE, training_batch_size
                ):
                    # zero the parameter gradients
                    optimizer.zero_grad()
                    # forward + backward + optimize
                    outputs = net(
                        inputs[
                            train_batch_start_idx : train_batch_start_idx
                            + training_batch_size
                        ].to(torch.float)
                    )
                    loss = criterion(
                        outputs,
                        labels[
                            train_batch_start_idx : train_batch_start_idx
                            + training_batch_size
                        ]
                        .to(torch.float)
                        .type(torch.LongTensor),
                    )
                    loss.backward()
                    optimizer.step()
                    if training_batch_idx % 100 == 0:
                        print(
                            "Train Epoch: {} Batch: {} Loss: {:.6f}".format(
                                epoch, training_batch_idx, loss.item()
                            )
                        )
                    training_batch_idx += 1

        model = PyTorchTileDBModel(
            uri="mnist_model",
            namespace=TILEDB_NAMESPACE,
            model=net,
            optimizer=optimizer,
        )

        # Save model as TileDB array.
        model.save()


tiledb.cloud.login(username=TILEDB_USER_NAME, password=TILEDB_PASSWD)

tiledb.cloud.udf.exec(train)

print(tiledb.cloud.last_udf_task().logs)



from math import ceil
from operator import itemgetter
from typing import Any, Iterable, Sequence, Union

import numpy as np

from .base import TensorSchema
from .ranges import ConstrainedPartitionsIntRange, InclusiveRange


class DenseTensorSchema(TensorSchema[np.ndarray]):
    """
    TensorSchema for reading dense TileDB arrays as (dense) Numpy arrays.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        if self._array.schema.sparse:
            raise NotImplementedError(
                "Mapping a sparse TileDB array to dense tensors is not implemented"
            )

    @property
    def key_range(self) -> ConstrainedPartitionsIntRange:
        self._key_range: ConstrainedPartitionsIntRange
        try:
            return self._key_range
        except AttributeError:
            key_dim_min, key_dim_max = self._ned[0]
            key_dim_slice = self._dim_selectors.get(0)
            if key_dim_slice is not None:
                assert isinstance(key_dim_slice, slice)
                min_key = key_dim_slice.start
                max_key = key_dim_slice.stop
            else:
                min_key = key_dim_min
                max_key = key_dim_max
            key_dim_tile = self._array.dim(self.key_dim).tile
            start_offsets = range(key_dim_min, key_dim_max + 1, key_dim_tile)
            self._key_range = ConstrainedPartitionsIntRange(
                min_key, max_key, start_offsets
            )
            return self._key_range

    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[int, int]]
    ) -> Union[Iterable[np.ndarray], Iterable[Sequence[np.ndarray]]]:
        """
        Generate batches of Numpy arrays.

        If `key_dim_index > 0`, the generated arrays will be reshaped so that the key_dim
        axes is first. For example, if the TileDB array `a` has shape (5, 12, 20) and
        `key_dim_index == 1`, then `a[:, 4:8, :]` returns arrays of shape (5, 4, 20) but
        this method yields arrays of shape (4, 5, 20).
        """
        query = self.query
        get_data = itemgetter(*self._fields)
        key_dim_index = self._key_dim_index
        for key_range in key_ranges:
            field_arrays = query[key_range.min : key_range.max]
            if key_dim_index > 0:
                # Move key_dim_index axes first
                for field, array in field_arrays.items():
                    field_arrays[field] = np.moveaxis(array, key_dim_index, 0)
            yield get_data(field_arrays)

    @property
    def max_partition_weight(self) -> int:
        memory_budget = int(self._array._ctx_().config()["sm.memory_budget"])

        # The memory budget should be large enough to read the cells of the largest field
        bytes_per_cell = max(dtype.itemsize for dtype in self.field_dtypes)

        # We want to be reading tiles following the tile extents along each dimension.
        # The number of cells for each such tile is the product of all tile extents.
        dim_tiles = [self._array.dim(dim).tile for dim in self._all_dims]
        cells_per_tile = np.prod(dim_tiles)

        # Each slice consists of `rows_per_slice` rows along the key dimension
        rows_per_slice = dim_tiles[0]

        # Reading a slice of `rows_per_slice` rows requires reading a number of tiles that
        # depends on the size, tile extent and selector of each dimension after the first one.
        shape = self.shape
        tiles_per_slice = 1
        for i, dim_tile in enumerate(dim_tiles[1:], 1):
            selector = self._dim_selectors.get(i)
            if selector is None:
                tiles_per_slice *= ceil(shape[i] / dim_tile)
            else:
                dim_start, dim_stop = self._ned[i]

                def get_tile_idx(value: int) -> int:
                    """Get the index of the tile with the given value in the i-th dimension"""
                    return int((value - dim_start) / dim_tile)

                if isinstance(selector, slice):
                    # count the number of tiles between start and stop (inclusive)
                    start = selector.start if selector.start is not None else dim_start
                    stop = selector.stop if selector.stop is not None else dim_stop
                    tiles_per_slice *= get_tile_idx(stop) - get_tile_idx(start) + 1
                else:
                    # count the number of unique tiles for the i-th dimension
                    tiles_per_slice *= len(set(map(get_tile_idx, selector)))

        # Compute the size in bytes of each slice of `rows_per_slice` rows
        bytes_per_slice = bytes_per_cell * cells_per_tile * tiles_per_slice

        # Compute the number of slices that fit within the memory budget
        num_slices = memory_budget // bytes_per_slice

        # Compute the total number of rows to slice
        return max(1, int(rows_per_slice * num_slices))



from operator import itemgetter
from typing import Any, Iterable, Optional, Sequence, Union

import numpy as np

from .base_sparse import BaseSparseTensorSchema
from .ranges import InclusiveRange

RaggedArray = Sequence[np.ndarray]


class RaggedTensorSchema(BaseSparseTensorSchema[RaggedArray]):
    """TensorSchema for reading sparse TileDB arrays as ragged Numpy arrays.

    Each ragged array is represented as a sequence of 1D Numpy arrays of the same dtype
    but (in general) different size. Each item of the ragged array contains the values of
    a given field for a given key (i.e. value along the key dimension). The items of the
    ragged array are ordered by the key dimension but the values within each item are not
    (meaningfully) ordered.

    Example:
    - (key, field1, field2) cells:
        {('a', 3, 2.8), ('b', 1, 0.2), ('a', 7, 3.2), ('c', 5, 6.1), ('c', 2, 0.5), ('a', 2, 1.9)}

    - Ragged arrays:
        - key: [np.array(['a', 'a', 'a']), np.array(['b']), np.array(['c', 'c'])]
        - field1: [np.array([3, 7, 2]), np.array([1]), np.array([5, 2])]
        - field2: [np.array([2.8, 3.2, 1.9]), np.array([0.2]), np.array([6.1, 0.5])]
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        if self.key_dim not in self._query_kwargs["dims"]:
            self._query_kwargs["dims"] += (self.key_dim,)

    @property
    def shape(self) -> Sequence[Optional[int]]:
        return len(self.key_range), None

    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[int, int]]
    ) -> Union[Iterable[RaggedArray], Iterable[Sequence[RaggedArray]]]:
        query = self.query
        get_data = itemgetter(*self._fields)
        key_dim = self.key_dim
        for key_range in key_ranges:
            field_arrays = query[key_range.min : key_range.max]
            # Sort the key dimension values and find the indices where the value changes
            sort_idx = np.argsort(field_arrays[key_dim], kind="stable")
            split_idx = argdiff(field_arrays[key_dim][sort_idx])
            # apply the same sorting and splitting to all the field arrays
            for name, values in field_arrays.items():
                field_arrays[name] = np.split(values[sort_idx], split_idx)
            yield get_data(field_arrays)


def argdiff(a: np.ndarray) -> np.ndarray:
    """Return the indices `i` of the array `a` so that `a[i] != a[i-1]`."""
    idx = np.nonzero(a[:-1] != a[1:])[0]
    # the minimum diff index (if a[0] != a[1]) is 1, not 0
    idx += 1
    return idx



"""Functionality for loading data from TileDB arrays to the Tensorflow Data API."""

from functools import singledispatch
from typing import Sequence, Union

import numpy as np
import scipy.sparse
import sparse
import tensorflow as tf

from ._tensor_schema import MappedTensorSchema, RaggedArray, SparseArray, TensorSchema
from .types import ArrayParams, TensorKind

Tensor = Union[np.ndarray, tf.SparseTensor]


def TensorflowTileDBDataset(
    *all_array_params: ArrayParams,
    num_workers: int = 0,
) -> tf.data.Dataset:
    """Return a tf.data.Dataset for loading data from TileDB arrays.
    :param all_array_params: One or more `ArrayParams` instances, one per TileDB array.
    :param num_workers: If greater than zero, create a threadpool of `num_workers` threads
        used to fetch inputs asynchronously and in parallel.
    """
    schemas = []
    for array_params in all_array_params:
        schema = array_params.tensor_schema
        if schema.kind is TensorKind.SPARSE_CSR:
            raise NotImplementedError(f"{schema.kind} tensors not supported")
        elif schema.kind is TensorKind.SPARSE_COO:
            schema = MappedTensorSchema(schema, _to_sparse_tensor)
        elif schema.kind is TensorKind.RAGGED:
            schema = MappedTensorSchema(schema, _to_ragged_tensor)
        schemas.append(schema)

    key_range = schemas[0].key_range
    if not all(key_range.equal_values(schema.key_range) for schema in schemas[1:]):
        raise ValueError(f"All arrays must have the same key range: {key_range}")

    max_weights = tuple(schema.max_partition_weight for schema in schemas)
    key_subranges = tuple(key_range.partition_by_count(num_workers or 1))

    def key_range_dataset(key_range_idx: int) -> tf.data.Dataset:
        datasets = tuple(
            tf.data.Dataset.from_generator(
                lambda i, schema=schema, max_weight=max_weight: schema.iter_tensors(
                    key_subranges[i].partition_by_weight(max_weight)
                ),
                args=(key_range_idx,),
                output_signature=_get_tensor_specs(schema),
            ).unbatch()
            for schema, max_weight in zip(schemas, max_weights)
        )
        return tf.data.Dataset.zip(datasets) if len(datasets) > 1 else datasets[0]

    if num_workers:
        dataset = tf.data.Dataset.from_tensor_slices(range(len(key_subranges)))
        dataset = dataset.interleave(
            key_range_dataset, num_parallel_calls=num_workers, deterministic=False
        )
    else:
        dataset = key_range_dataset(0)

    return dataset


_tensor_specs = {
    TensorKind.DENSE: tf.TensorSpec,
    TensorKind.SPARSE_COO: tf.SparseTensorSpec,
    TensorKind.RAGGED: tf.RaggedTensorSpec,
}


def _get_tensor_specs(
    schema: TensorSchema[Tensor],
) -> Union[tf.TypeSpec, Sequence[tf.TypeSpec]]:
    spec_cls = _tensor_specs[schema.kind]
    shape = (None, *schema.shape[1:])
    specs = tuple(spec_cls(shape=shape, dtype=dtype) for dtype in schema.field_dtypes)
    return specs if len(specs) > 1 else specs[0]


@singledispatch
def _to_sparse_tensor(sa: SparseArray) -> tf.SparseTensor:
    """Create a tf.SparseTensor from a sparse array"""
    raise NotImplementedError


@_to_sparse_tensor.register(scipy.sparse.csr_matrix)
def _csr_to_sparse_tensor(csr: scipy.sparse.csr_matrix) -> tf.SparseTensor:
    """Create a tf.SparseTensor from a scipy.sparse.csr_matrix instance"""
    coo = csr.tocoo()
    coords = np.array((coo.row, coo.col))
    return tf.SparseTensor(coords.T, coo.data, coo.shape)


@_to_sparse_tensor.register(sparse.COO)
def _coo_to_sparse_tensor(coo: sparse.COO) -> tf.SparseTensor:
    """Create a tf.SparseTensor from a sparse.COO instance"""
    return tf.SparseTensor(coo.coords.T, coo.data, coo.shape)


def _to_ragged_tensor(ra: RaggedArray) -> tf.RaggedTensor:
    return tf.ragged.constant(ra, dtype=ra[0].dtype)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np

import tiledb

from ..types import Selector, TensorKind
from .query import KeyDimQuery
from .ranges import InclusiveRange

Tensor = TypeVar("Tensor")


@dataclass(frozen=True)
class TensorSchema(ABC, Generic[Tensor]):
    """
    A class to encapsulate the information needed for mapping a TileDB array to tensors.
    """

    kind: TensorKind
    _array: tiledb.Array
    _key_dim_index: int
    _fields: Sequence[str]
    _all_dims: Sequence[str]
    _ned: Sequence[Tuple[Any, Any]]
    _dim_selectors: Dict[int, Selector]
    _query_kwargs: Dict[str, Any]

    @property
    def num_fields(self) -> int:
        """Number of attributes and dimensions to read."""
        return len(self._fields)

    @property
    def field_dtypes(self) -> Sequence[np.dtype]:
        """Dtypes of attributes and dimensions to read."""
        return tuple(map(self._array.schema.attr_or_dim_dtype, self._fields))

    @property
    def shape(self) -> Sequence[Optional[int]]:
        """Shape of the selected array, with the key dimension moved first.

        The size of each dimension is determined by:
        - either the respective selector in `_dim_selectors` (if given)
        - or by the non-empty domain otherwise.

        :raises ValueError: If the array does not have integer domain.
        """
        shape = [len(self.key_range)]
        for i, (dim_start, dim_stop) in enumerate(self._ned[1:], 1):
            selector = self._dim_selectors.get(i)
            if selector is None:
                selector = slice(None)
            if isinstance(selector, slice):
                start = selector.start if selector.start is not None else dim_start
                stop = selector.stop if selector.stop is not None else dim_stop
                if not (isinstance(start, int) and isinstance(stop, int)):
                    raise ValueError("Shape not defined for non-integer domain")
                shape.append(stop - start + 1)
            else:
                shape.append(len(selector))
        return tuple(shape)

    @property
    def query(self) -> KeyDimQuery:
        """A sliceable object for querying the TileDB array along the key dimension"""
        return self._get_query(**self._query_kwargs)

    @property
    def key_dim(self) -> str:
        """Key dimension of the array."""
        return self._all_dims[0]

    @property
    @abstractmethod
    def key_range(self) -> InclusiveRange[Any, int]:
        """Inclusive range of the key dimension.

        The values of the range are all the distinct values of the key dimension (keys).
        The weight of each key is:
        - for dense arrays: 1
        - for sparse arrays: The number of non-empty cells for this key
        """

    @property
    @abstractmethod
    def max_partition_weight(self) -> int:
        """
        Determine the maximum partition that can be read without incomplete retries.

        What constitutes weight of a partition depends on the array type:
        - For dense arrays, it is the number of unique keys (= number of "rows").
          It depends on the `sm.memory_budget` config parameter.
        - For sparse arrays, it is the number of non-empty cells.
          It depends on the `py.init_buffer_bytes` config parameter.
        """

    @abstractmethod
    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[Any, int]]
    ) -> Union[Iterable[Tensor], Iterable[Sequence[Tensor]]]:
        """
        Generate batches of tensors.

        Each yielded batch is either:
        - a sequence of N tensors if N > 1, where `N == self.num_fields`, or
        - a single tensor if N == 1.
        where each tensor has shape `(len(key_range), *self.shape[1:])`.

        :param key_ranges: Inclusive ranges along the key dimension.
        """

    def _get_query(self, **kwargs: Any) -> KeyDimQuery:
        return KeyDimQuery(
            self._array, self._key_dim_index, self._dim_selectors, **kwargs
        )



from typing import Any, Callable, Generic, Iterable, Sequence, TypeVar, Union

import wrapt

from .base import Tensor, TensorSchema
from .ranges import InclusiveRange

MappedTensor = TypeVar("MappedTensor")


class MappedTensorSchema(wrapt.ObjectProxy, Generic[Tensor, MappedTensor]):
    """
    Proxy class that wraps a TensorSchema and applies a mapping function to each tensor
    yielded by `iter_tensors`.
    """

    def __init__(
        self,
        wrapped: TensorSchema[Tensor],
        map_tensor: Callable[[Tensor], MappedTensor],
    ):
        super().__init__(wrapped)
        self._self_map_tensor = map_tensor

    def iter_tensors(
        self, key_ranges: Iterable[InclusiveRange[Any, int]]
    ) -> Union[Iterable[MappedTensor], Iterable[Sequence[MappedTensor]]]:
        wrapped_iter_tensors = self.__wrapped__.iter_tensors(key_ranges)
        if self.num_fields == 1:
            return map(self._self_map_tensor, wrapped_iter_tensors)
        else:
            return (
                tuple(map(self._self_map_tensor, tensors))
                for tensors in wrapped_iter_tensors
            )

    def __reduce_ex__(self, protocol):  # type: ignore
        return MappedTensorSchema, (self.__wrapped__, self._self_map_tensor)



"""Functionality for loading data from TileDB arrays to the PyTorch Dataloader API."""

from functools import partial
from typing import Any, Callable, Iterator, Sequence, Tuple, Union

import numpy as np
import scipy.sparse
import sparse
import torchdata
from torch.utils.data import DataLoader, IterDataPipe
from torchdata.datapipes.iter import IterableWrapper

from ._pytorch_collators import Collator
from ._tensor_schema import TensorSchema
from ._tensor_schema.ranges import InclusiveRange
from .types import ArrayParams, TensorKind

TensorLike = Union[np.ndarray, sparse.COO, scipy.sparse.csr_matrix]
TensorLikeOrTuple = Union[TensorLike, Tuple[TensorLike, ...]]


def PyTorchTileDBDataLoader(
    *all_array_params: ArrayParams,
    shuffle_buffer_size: int = 0,
    **kwargs: Any,
) -> DataLoader:
    """Return a DataLoader for loading data from TileDB arrays.

    :param all_array_params: One or more `ArrayParams` instances, one per TileDB array.
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    **kwargs: Should contain all parameters for PyTorch Dataloader. At the moment TileDB-ML can support ONLY the
    following PyTorch Dataloader arguments:
        batch_size: How many samples per batch to load (default: ``1``).
        prefetch_factor: Number of batches loaded in advance by each worker. Not applicable (and should not be
        given) when `num_workers` is 0.
        num_workers: How many subprocesses to use for data loading. 0 means that the data will be loaded in the main
        process. Note: when `num_workers` > 1 yielded batches may be shuffled even if `shuffle_buffer_size` is zero.
        persistent_workers: If ``True``, the data loader will not shutdown the worker processes after a dataset has
        been consumed once. This allows to maintain the workers `Dataset` instances alive. (default: ``False``)
        timeout: if positive, the timeout value for collecting a batch from workers. Should always be non-negative.
        (default: ``0``)
        drop_last: Set to ``True`` to drop the last incomplete batch, if the dataset size is not divisible by the
        batch size. If ``False`` and the size of dataset is not divisible by the batch size, then the last batch
        will be smaller. (default: ``False``)

    Users should NOT pass (TileDB-ML either doesn't support or implements internally the corresponding functionality)
    the following arguments: 'shuffle', 'sampler', 'batch_sampler', 'worker_init_fn' and 'collate_fn'.
    """
    is_batched = kwargs.get("batch_size", 1) is not None

    schemas = []
    for array_params in all_array_params:
        schema = array_params.tensor_schema
        # unbatched 3D arrays generate 2D tensors so they can be converted to CSR
        # anything else with ndim>=3 cannot
        if schema.kind is TensorKind.SPARSE_CSR:
            ndim = len(schema.shape)
            if ndim > 3 or ndim == 3 and is_batched:
                raise ValueError(f"Cannot generate CSR tensors for {ndim}D array")
        schemas.append(schema)

    key_range = schemas[0].key_range
    if not all(key_range.equal_values(schema.key_range) for schema in schemas[1:]):
        raise ValueError(f"All arrays must have the same key range: {key_range}")

    datapipe_for_key_range = partial(_get_unbatched_datapipe, schemas)
    num_workers = kwargs.get("num_workers", 0)
    if num_workers:
        if torchdata.__version__ < "0.4":
            raise NotImplementedError("torchdata>=0.4 required for multiple workers")
        if any(schema.kind is not TensorKind.DENSE for schema in schemas):
            raise NotImplementedError("https://github.com/pytorch/pytorch/issues/20248")

        # partition the key range into `num_workers` subkey ranges of roughly equal weight
        worker_key_ranges = tuple(key_range.partition_by_count(num_workers))
        # create a datapipe for these partitions
        datapipe = IterableWrapper(worker_key_ranges, deepcopy=False)
        # shard the datapipe so that each worker gets exactly one partition
        datapipe = datapipe.sharding_filter()
        # read and unbatch the tensors for each partition
        datapipe = datapipe.flatmap(datapipe_for_key_range)
    else:
        # create a datapipe that reads and unbatches the tensors for the whole key range
        datapipe = datapipe_for_key_range(key_range)

    # shuffle the unbatched rows if shuffle_buffer_size > 0
    if shuffle_buffer_size:
        # load the rows to be shuffled
        # don't batch them (batch_size=None) or collate them (collate_fn=_identity)
        row_loader = DataLoader(
            datapipe, num_workers=num_workers, batch_size=None, collate_fn=_identity
        )
        # create a new datapipe for these rows
        datapipe = DeferredIterableIterDataPipe(iter, row_loader)
        # shuffle the datapipe items
        datapipe = datapipe.shuffle(buffer_size=shuffle_buffer_size)
        # run the shuffling on this process, not on workers
        kwargs["num_workers"] = 0

    # construct an appropriate collate function
    collator = Collator.from_schemas(*schemas)
    kwargs["collate_fn"] = collator.collate if is_batched else collator.convert

    # return the DataLoader for the final datapipe
    return DataLoader(datapipe, **kwargs)


class DeferredIterableIterDataPipe(IterDataPipe):
    """Wraps a callable that returns an iterable object to create an IterDataPipe."""

    def __init__(self, func: Callable[..., Iterator[Any]], *args: Any, **kwargs: Any):
        self._callable = partial(func, *args, **kwargs)

    def __iter__(self) -> Iterator[Any]:
        return self._callable()


def _identity(x: Any) -> Any:
    return x


def _get_unbatched_datapipe(
    schemas: Sequence[TensorSchema[TensorLike]],
    key_range: InclusiveRange[Any, int],
) -> IterDataPipe[Union[TensorLikeOrTuple, Tuple[TensorLikeOrTuple, ...]]]:
    """Return a datapipe over unbatched rows for the given schemas and key range.

    If `len(schemas) == 1`, each item of the datapipe is either a single `TensorLike`
    or a sequence of `TensorLike`s, depending on `schemas[0].num_fields`.
    If `len(schemas) > 1`, each item of the datapipe is a tuple of (`TensorLike` or
    sequence of `TensorLike`s, depending on `schema.num_fields`), one for each schema.
    """
    schema_dps = [
        DeferredIterableIterDataPipe(_unbatch_tensors, schema, key_range)
        for schema in schemas
    ]
    dp = schema_dps.pop(0)
    if schema_dps:
        dp = dp.zip(*schema_dps)
    return dp


def _unbatch_tensors(
    schema: TensorSchema[TensorLike], key_range: InclusiveRange[Any, int]
) -> Iterator[TensorLikeOrTuple]:
    """
    Generate batches of `TensorLike`s for the given schema and key range and then unbatch
    them into single "rows".

    If `schema.num_fields == 1`, each "row" is a single `TensorLike`
    If `schema.num_fields > 1`, each "row" is a sequence of `TensorLike`s
    """
    batches = schema.iter_tensors(
        key_range.partition_by_weight(schema.max_partition_weight)
    )
    if schema.num_fields > 1:
        # convert batches of columns to batches of rows
        batches = (zip(*batch) for batch in batches)
    # flatten batches of rows
    return (row for batch in batches for row in batch)



import numpy as np
import pytest
import scipy.sparse
import sparse
import torch

from tiledb.ml.readers import _pytorch_collators as pc

from .utils import get_tensor_kind


class TestNumpyArrayCollator:
    def test_convert_dense(self):
        value = np.random.rand(2, 5)
        tensor = pc.NumpyArrayCollator().convert(value)
        assert tensor.shape == (2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.DENSE
        np.testing.assert_array_equal(tensor, value)

    def test_collate_dense(self):
        batch = [np.random.rand(2, 5) for _ in range(7)]
        tensor = pc.NumpyArrayCollator().collate(batch)
        assert tensor.shape == (7, 2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.DENSE
        np.testing.assert_array_equal(tensor, np.stack(batch))

    def test_convert_nested(self):
        value = np.random.rand(10)
        tensor = pc.NumpyArrayCollator(to_nested=True).convert(value)
        assert tensor.shape == (10,)
        assert get_tensor_kind(tensor) is pc.TensorKind.DENSE
        np.testing.assert_array_equal(tensor, value)

    @pytest.mark.skipif(
        not hasattr(torch, "nested"), reason="Nested tensors not supported"
    )
    def test_collate_nested(self):
        batch = [np.random.rand(8), np.random.rand(3), np.random.rand(7)]
        tensor = pc.NumpyArrayCollator(to_nested=True).collate(batch)
        assert list(t.shape for t in tensor) == [(8,), (3,), (7,)]
        assert get_tensor_kind(tensor) is pc.TensorKind.RAGGED
        for tensor_row, batch_row in zip(tensor, batch):
            np.testing.assert_array_equal(tensor_row, batch_row)


class TestSparseCOOCollator:
    def test_convert_to_coo(self):
        value = sparse.random((2, 5), nnz=3)
        tensor = pc.SparseCOOCollator().convert(value)
        assert tensor.shape == (2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(tensor.to_dense(), value.todense())

    def test_collate_to_coo(self):
        batch = [sparse.random((2, 5), nnz=nnz) for nnz in np.random.randint(0, 11, 7)]
        tensor = pc.SparseCOOCollator().collate(batch)
        assert tensor.shape == (7, 2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(tensor.to_dense(), sparse.stack(batch).todense())

    def test_convert_to_csr(self):
        value = sparse.random((2, 5), nnz=3)
        tensor = pc.SparseCOOCollator(to_csr=True).convert(value)
        assert tensor.shape == (2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_CSR
        np.testing.assert_array_equal(tensor.to_dense(), value.todense())

    def test_collate_to_csr(self):
        batch = [sparse.random(10, nnz=nnz) for nnz in np.random.randint(0, 11, 7)]
        tensor = pc.SparseCOOCollator(to_csr=True).collate(batch)
        assert tensor.shape == (7, 10)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_CSR
        np.testing.assert_array_equal(tensor.to_dense(), sparse.stack(batch).todense())


class TestScipySparseCSRCollator:
    def test_convert_to_coo(self):
        value = scipy.sparse.random(2, 5, density=0.3, format="csr")
        tensor = pc.ScipySparseCSRCollator().convert(value)
        assert tensor.shape == (2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(tensor.to_dense(), value.toarray())

    def test_convert_to_coo_single_row(self):
        value = scipy.sparse.random(1, 10, density=0.3, format="csr")
        tensor = pc.ScipySparseCSRCollator().convert(value)
        assert tensor.shape == (10,)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(tensor.to_dense(), value.toarray()[0])

    def test_collate_to_coo(self):
        batch = [
            scipy.sparse.random(1, 10, density=nnz / 10, format="csr")
            for nnz in np.random.randint(0, 11, 7)
        ]
        tensor = pc.ScipySparseCSRCollator().collate(batch)
        assert tensor.shape == (7, 10)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(
            tensor.to_dense(), scipy.sparse.vstack(batch).toarray()
        )

    def test_convert_to_csr(self):
        value = scipy.sparse.random(2, 5, density=0.3, format="csr")
        tensor = pc.ScipySparseCSRCollator(to_csr=True).convert(value)
        assert tensor.shape == (2, 5)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_CSR
        np.testing.assert_array_equal(tensor.to_dense(), value.toarray())

    def test_collate_to_csr(self):
        batch = [
            scipy.sparse.random(1, 10, density=nnz / 10, format="csr")
            for nnz in np.random.randint(0, 11, 7)
        ]
        tensor = pc.ScipySparseCSRCollator(to_csr=True).collate(batch)
        assert tensor.shape == (7, 10)
        assert get_tensor_kind(tensor) is pc.TensorKind.SPARSE_CSR
        np.testing.assert_array_equal(
            tensor.to_dense(), scipy.sparse.vstack(batch).toarray()
        )


class TestRowCollator:
    collator = pc.RowCollator(
        [
            pc.NumpyArrayCollator(),
            pc.SparseCOOCollator(),
            pc.ScipySparseCSRCollator(to_csr=True),
        ]
    )

    def test_convert(self):
        row = self._get_row()
        tensors = self.collator.convert(row)

        assert isinstance(tensors, tuple)
        assert len(tensors) == 3

        assert tensors[0].shape == (10,)
        assert get_tensor_kind(tensors[0]) is pc.TensorKind.DENSE
        np.testing.assert_array_equal(tensors[0], row[0])

        assert tensors[1].shape == (3, 4, 2)
        assert get_tensor_kind(tensors[1]) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(tensors[1].to_dense(), row[1].todense())

        assert tensors[2].shape == (1, 5)
        assert get_tensor_kind(tensors[2]) is pc.TensorKind.SPARSE_CSR
        np.testing.assert_array_equal(tensors[2].to_dense(), row[2].toarray())

    def test_collate(self):
        rows = [self._get_row() for _ in range(7)]
        tensors = self.collator.collate(rows)
        columns = list(zip(*rows))

        assert isinstance(tensors, tuple)
        assert len(tensors) == 3

        assert tensors[0].shape == (7, 10)
        assert get_tensor_kind(tensors[0]) is pc.TensorKind.DENSE
        np.testing.assert_array_equal(tensors[0], columns[0])

        assert tensors[1].shape == (7, 3, 4, 2)
        assert get_tensor_kind(tensors[1]) is pc.TensorKind.SPARSE_COO
        np.testing.assert_array_equal(
            tensors[1].to_dense(), sparse.stack(columns[1]).todense()
        )

        assert tensors[2].shape == (7, 5)
        assert get_tensor_kind(tensors[2]) is pc.TensorKind.SPARSE_CSR
        np.testing.assert_array_equal(
            tensors[2].to_dense(), scipy.sparse.vstack(columns[2]).toarray()
        )

    @staticmethod
    def _get_row():
        return (
            np.random.rand(10),
            sparse.random((3, 4, 2), nnz=np.random.randint(24)),
            scipy.sparse.random(1, 5, density=np.random.randint(10) / 10, format="csr"),
        )


class Test_Collator_from_schemas:
    @pytest.mark.parametrize(
        "tensor_kind,to_nested",
        [(pc.TensorKind.DENSE, False), (pc.TensorKind.RAGGED, True)],
    )
    def test_numpy(self, mocker, tensor_kind, to_nested):
        schema = mocker.Mock(kind=tensor_kind, num_fields=1)
        collator = pc.Collator.from_schemas(schema)
        assert isinstance(collator, pc.NumpyArrayCollator)
        assert collator.to_nested is to_nested
        self._test_multiple_fields(schema, collator)

    @pytest.mark.parametrize(
        "tensor_kind,shape,collator_cls,to_csr",
        [
            (pc.TensorKind.SPARSE_COO, (3, 4, 2), pc.SparseCOOCollator, False),
            (pc.TensorKind.SPARSE_CSR, (3, 4, 2), pc.SparseCOOCollator, True),
            (pc.TensorKind.SPARSE_COO, (3, 4), pc.ScipySparseCSRCollator, False),
            (pc.TensorKind.SPARSE_CSR, (3, 4), pc.ScipySparseCSRCollator, True),
        ],
    )
    def test_sparse(self, mocker, tensor_kind, shape, collator_cls, to_csr):
        schema = mocker.Mock(kind=tensor_kind, shape=shape, num_fields=1)
        collator = pc.Collator.from_schemas(schema)
        assert isinstance(collator, collator_cls)
        assert collator.to_csr is to_csr
        self._test_multiple_fields(schema, collator)

    def test_multiple(self, mocker):
        schemas = (
            mocker.Mock(kind=pc.TensorKind.DENSE, num_fields=1),
            mocker.Mock(kind=pc.TensorKind.SPARSE_CSR, shape=(3, 4, 2), num_fields=2),
            mocker.Mock(kind=pc.TensorKind.SPARSE_COO, shape=(3, 4), num_fields=3),
        )

        multi_collator = pc.Collator.from_schemas(*schemas)
        assert isinstance(multi_collator, pc.RowCollator)
        assert len(multi_collator.column_collators) == 3

        collator = multi_collator.column_collators[0]
        assert isinstance(collator, pc.NumpyArrayCollator)
        assert collator.to_nested is False

        collator = multi_collator.column_collators[1]
        assert isinstance(collator, pc.RowCollator)
        assert len(collator.column_collators) == 2
        for sub_collator in collator.column_collators:
            assert isinstance(sub_collator, pc.SparseCOOCollator)
            assert sub_collator.to_csr is True

        collator = multi_collator.column_collators[2]
        assert isinstance(collator, pc.RowCollator)
        assert len(collator.column_collators) == 3
        for sub_collator in collator.column_collators:
            assert isinstance(sub_collator, pc.ScipySparseCSRCollator)
            assert sub_collator.to_csr is False

    @staticmethod
    def _test_multiple_fields(schema, collator):
        schema.num_fields = 2
        row_collator = pc.Collator.from_schemas(schema)
        assert isinstance(row_collator, pc.RowCollator)
        assert row_collator.column_collators == (collator, collator)



"""Functionality for saving and loading Tensorflow Keras models as TileDB arrays"""

import io
import json
import logging
import os
import pickle
from collections import ChainMap
from typing import Any, Mapping, Optional, Tuple

import keras
import tensorflow as tf

import tiledb

from ._base import Meta, TileDBArtifact, Timestamp

SharedObjectLoadingScope = keras.utils.generic_utils.SharedObjectLoadingScope
FunctionalOrSequential = (keras.models.Functional, keras.models.Sequential)
TFOptimizer = keras.optimizers.TFOptimizer
get_json_type = keras.saving.saved_model.json_utils.get_json_type
preprocess_weights_for_loading = keras.saving.hdf5_format.preprocess_weights_for_loading
saving_utils = keras.saving.saving_utils


class TensorflowKerasTileDBModel(TileDBArtifact[tf.keras.Model]):
    """
    Class that implements all functionality needed to save Tensorflow models as
    TileDB arrays and load Tensorflow models from TileDB arrays.
    """

    Name = "TENSORFLOW KERAS"
    Version = tf.__version__

    def __init__(
        self,
        uri: str,
        namespace: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        model: Optional[tf.keras.Model] = None,
    ):
        super().__init__(uri, namespace, ctx, model)

    def save(
        self,
        *,
        meta: Optional[Meta] = None,
        include_optimizer: bool = False,
        callbacks: Optional[tf.keras.callbacks.CallbackList] = None,
    ) -> None:
        """
        Save a Tensorflow model as a TileDB array.

        :param meta: Extra metadata to save in a TileDB array.
        :param include_optimizer: Whether to save the optimizer or not.
        :param callbacks: Callbacks list to store. At the moment only Tensorboard callback is supported.
        """
        if self.artifact is None:
            raise RuntimeError("Model is not initialized")

        if not isinstance(self.artifact, FunctionalOrSequential):
            raise RuntimeError(
                "Subclassed Models (Custom Layers) not supported at the moment."
            )

        # Used in this format only when model is Functional or Sequential
        model_weights = pickle.dumps(self.artifact.get_weights(), protocol=4)

        # Serialize model optimizer
        if include_optimizer:
            optimizer_weights = self._serialize_optimizer_weights()
        else:
            optimizer_weights = b""

        # Serialize Tensorboard files
        tensorboard_log_dir = None
        if callbacks:
            for cb in callbacks:
                if isinstance(cb, tf.keras.callbacks.TensorBoard):
                    tensorboard_log_dir = os.path.join(cb.log_dir, "train")
                else:
                    raise NotImplementedError(cb)

        # Create TileDB model array
        if not tiledb.array_exists(uri=self.uri):
            self._create_array(fields=["model", "optimizer", "tensorboard"])

        # Write extra metadata. Only for Tensoflow models.
        model_meta = saving_utils.model_metadata(
            model=self.artifact,
            include_optimizer=bool(optimizer_weights),
        )
        for key, value in model_meta.items():
            model_meta[key] = json.dumps(value, default=get_json_type).encode("utf8")

        self._write_array(
            model_params={
                "model": model_weights,
                "optimizer": optimizer_weights,
            },
            tensorboard_log_dir=tensorboard_log_dir,
            meta=ChainMap(model_meta, dict(meta or ())),
        )

    def load(
        self,
        *,
        timestamp: Optional[Timestamp] = None,
        compile_model: bool = False,
        custom_objects: Optional[Mapping[str, Any]] = None,
        input_shape: Optional[Tuple[int, ...]] = None,
        callback: bool = False,
    ) -> tf.keras.Model:
        """
        Load switch, i.e, decide between __load (TileDB-ML<=0.8.0) or __load_v2 (TileDB-ML>0.8.0).

        :param callback: Boolean variable if True will store callback data into saved directory. At the moment supports
        only Tensorboard callback.
        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        :param compile_model: Whether to compile the model after loading or not.
        :param custom_objects: Mapping of names to custom classes or functions to be
            considered during deserialization.
        :param input_shape: The shape that the custom model expects as input
        :return: Tensorflow model.
        """
        with tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp) as model_array:
            if self._use_legacy_schema(model_array):
                return self.__load_legacy(
                    model_array, compile_model, callback, custom_objects
                )
            else:
                return self.__load(model_array, compile_model, callback)

    def __load_legacy(
        self,
        model_array: tiledb.Array,
        compile_model: bool,
        callback: bool,
        custom_objects: Optional[Mapping[str, Any]],
    ) -> tf.keras.Model:
        model_array_results = model_array[:]
        model_config = json.loads(model_array.meta["model_config"])
        model_class = model_config["class_name"]

        cls = tf.keras.Sequential if model_class == "Sequential" else tf.keras.Model
        model = cls.from_config(model_config["config"])
        model_weights = pickle.loads(model_array_results["model_weights"].item(0))
        model.set_weights(model_weights)

        if compile_model:
            optimizer_weights = pickle.loads(
                model_array_results["optimizer_weights"].item(0)
            )
            training_config = json.loads(model_array.meta["training_config"])

            # Compile model.
            model.compile(
                **saving_utils.compile_args_from_training_config(
                    training_config, custom_objects
                )
            )
            saving_utils.try_build_compiled_arguments(model)

            # Set optimizer weights.
            if optimizer_weights:
                try:
                    model.optimizer._create_all_weights(model.trainable_variables)
                except (NotImplementedError, AttributeError):
                    logging.warning(
                        "Error when creating the weights of optimizer {}, making it "
                        "impossible to restore the saved optimizer state. As a result, "
                        "your model is starting with a freshly initialized optimizer."
                    )

                try:
                    model.optimizer.set_weights(optimizer_weights)
                except ValueError:
                    logging.warning(
                        "Error in loading the saved optimizer "
                        "state. As a result, your model is "
                        "starting with a freshly initialized "
                        "optimizer."
                    )

        if callback:
            try:
                with tiledb.open(f"{self.uri}-tensorboard") as tb_array:
                    for path, file_bytes in pickle.loads(
                        tb_array[:]["tensorboard_data"][0]
                    ).items():
                        log_dir = os.path.dirname(path)
                        if not os.path.exists(log_dir):
                            os.mkdir(log_dir)
                        with open(
                            os.path.join(log_dir, os.path.basename(path)), "wb"
                        ) as f:
                            f.write(file_bytes)
            except FileNotFoundError:
                print(f"Array {self.uri}-tensorboard does not exist")
        return model

    def __load(
        self, model_array: tiledb.Array, compile_model: bool, callback: bool
    ) -> tf.keras.Model:
        model_config = json.loads(model_array.meta["model_config"])
        model_class = model_config["class_name"]

        cls = tf.keras.Sequential if model_class == "Sequential" else tf.keras.Model
        model = cls.from_config(model_config["config"])
        model_weights = self._get_model_param(model_array, "model")
        model.set_weights(model_weights)

        if compile_model:
            training_config = json.loads(model_array.meta["training_config"])

            # Compile model.
            model.compile(
                **saving_utils.compile_args_from_training_config(
                    training_config,
                )
            )

            saving_utils.try_build_compiled_arguments(model)

            optimizer_weights = self._get_model_param(model_array, "optimizer")

            # Set optimizer weights.
            if optimizer_weights:
                try:
                    model.optimizer._create_all_weights(model.trainable_variables)
                except (NotImplementedError, AttributeError):
                    logging.warning(
                        "Error when creating the weights of optimizer {}, making it "
                        "impossible to restore the saved optimizer state. As a result, "
                        "your model is starting with a freshly initialized optimizer."
                    )

                try:
                    model.optimizer.set_weights(optimizer_weights)
                except ValueError:
                    logging.warning(
                        "Error in loading the saved optimizer "
                        "state. As a result, your model is "
                        "starting with a freshly initialized "
                        "optimizer."
                    )

        if callback:
            self._load_tensorboard(model_array)

        return model

    def preview(self) -> str:
        """Create a string representation of the Tensorflow model."""
        if self.artifact:
            str_rep = io.StringIO()
            self.artifact.summary(print_fn=lambda x: str_rep.write(x + "\n"))
            model_summary = str_rep.getvalue()
            return model_summary
        return ""

    def _serialize_optimizer_weights(
        self,
    ) -> bytes:
        """Serialize optimizer weights."""
        assert self.artifact
        optimizer = self.artifact.optimizer
        if optimizer and not isinstance(optimizer, TFOptimizer):
            optimizer_weights = tf.keras.backend.batch_get_value(optimizer.weights)
            return pickle.dumps(optimizer_weights, protocol=4)
        return b""



from math import ceil
from typing import Any, Counter

import numpy as np

from .base import Tensor, TensorSchema
from .ranges import WeightedRange


class BaseSparseTensorSchema(TensorSchema[Tensor]):
    """Abstract base class for reading sparse TileDB arrays."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        if not self._array.schema.sparse:
            raise NotImplementedError(
                "Mapping a dense TileDB array to sparse tensors is not implemented"
            )

    @property
    def key_range(self) -> WeightedRange[Any, int]:
        self._key_range: WeightedRange[Any, int]
        try:
            return self._key_range
        except AttributeError:
            key_counter = Counter[Any]()
            key_dim = self.key_dim
            query = self._get_query(dims=(key_dim,), attrs=(), return_incomplete=True)
            key_dim_slice = self._dim_selectors.get(0, slice(None))
            assert isinstance(key_dim_slice, slice)
            for result in query[key_dim_slice]:
                key_counter.update(result[key_dim])
            self._key_range = WeightedRange.from_mapping(key_counter)
            return self._key_range

    @property
    def max_partition_weight(self) -> int:
        try:
            memory_budget = int(self._array._ctx_().config()["py.init_buffer_bytes"])
        except KeyError:
            memory_budget = 10 * 1024**2

        # Determine the bytes per (non-empty) cell for each field.
        # - For fixed size fields, this is just the `dtype.itemsize` of the field.
        # - For variable size fields, the best we can do is to estimate the average bytes
        #   size. We also need to take into account the (fixed) offset buffer size per
        #   cell (=8 bytes).
        offset_itemsize = np.dtype(np.uint64).itemsize
        attr_or_dim_dtype = self._array.schema.attr_or_dim_dtype
        query = self._get_query(return_incomplete=True, **self._query_kwargs)
        bytes_per_cell = []
        for field, est_result_size in query[:].estimated_result_sizes().items():
            if est_result_size.offsets_bytes == 0:
                bytes_per_cell.append(attr_or_dim_dtype(field).itemsize)
            else:
                num_cells = est_result_size.offsets_bytes / offset_itemsize
                avg_itemsize = est_result_size.data_bytes / num_cells
                bytes_per_cell.append(max(avg_itemsize, offset_itemsize))

        # Finally, the number of cells that can fit in the memory_budget depends on the
        # maximum bytes_per_cell
        return max(1, memory_budget // ceil(max(bytes_per_cell)))



from enum import Enum, unique


@unique
class ModelFileProperties(Enum):
    """
    Enum Class that contains all model array file properties.
    """

    TILEDB_ML_MODEL_ML_FRAMEWORK = "TILEDB_ML_MODEL_ML_FRAMEWORK"
    TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION = "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"
    TILEDB_ML_MODEL_STAGE = "TILEDB_ML_MODEL_STAGE"
    TILEDB_ML_MODEL_PYTHON_VERSION = "TILEDB_ML_MODEL_PYTHON_VERSION"
    TILEDB_ML_MODEL_PREVIEW = "TILEDB_ML_MODEL_PREVIEW"
    TILEDB_ML_MODEL_VERSION = "TILEDB_ML_MODEL_VERSION"



import enum
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence, Union

import numpy as np

import tiledb

if TYPE_CHECKING:
    from ._tensor_schema import TensorSchema

Selector = Union[slice, Sequence[int]]


class TensorKind(enum.Enum):
    """Kind of tensor."""

    DENSE = enum.auto()
    SPARSE_COO = enum.auto()
    SPARSE_CSR = enum.auto()
    RAGGED = enum.auto()


@dataclass(frozen=True)
class ArrayParams:
    """
    Data class specifying the parameters for loading tensors from a TileDB array.

    Public attributes:
    - array: TileDB array to be accessed
    - key_dim: Name (or index) of the array key dimension. Defaults to the first dimension.
    - fields: Fields (dimensions and attributes) to be retrieved from array. Defaults to
      all attributes of the array.
    - dim_selectors: Mapping from dimension name to a slice or sequence of indices of this
      dimension to select.
    - tensor_kind: kind of tensor desired. If not specified, it is determined based on the
      array schema.
    """

    array: tiledb.Array
    key_dim: Union[int, str] = 0
    fields: Sequence[str] = ()
    dim_selectors: Mapping[str, Selector] = field(default_factory=dict)
    tensor_kind: Optional[TensorKind] = None
    _tensor_schema_kwargs: Mapping[str, Any] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        all_attrs = [self.array.attr(i).name for i in range(self.array.nattr)]
        all_dims = [self.array.dim(i).name for i in range(self.array.ndim)]
        dims = []

        if self.fields:
            attrs = []
            for f in self.fields:
                if f in all_attrs:
                    attrs.append(f)
                elif f in all_dims:
                    dims.append(f)
                else:
                    raise ValueError(f"Unknown attribute or dimension '{f}'")
            final_fields = self.fields
        else:
            final_fields = attrs = all_attrs

        ned = list(self.array.nonempty_domain())
        key_dim_index = (
            self.key_dim
            if isinstance(self.key_dim, int)
            else all_dims.index(self.key_dim)
        )

        if key_dim_index > 0:
            # Swap key dimension to first position
            all_dims[0], all_dims[key_dim_index] = all_dims[key_dim_index], all_dims[0]
            ned[0], ned[key_dim_index] = ned[key_dim_index], ned[0]

        dim_selector_indices = {}
        for dim, selector in self.dim_selectors.items():
            i = all_dims.index(dim)
            if not isinstance(selector, (slice, Sequence)):
                raise TypeError("dim_selectors values must be slices or sequences")
            if i == 0:
                # key dimension selector
                if not isinstance(selector, slice):
                    raise TypeError("Key dimension selector must be a slice")
                if selector.step is not None:
                    raise ValueError("Stepped slice ranges are not supported")
                # intersect the non-empty domain of the key dimension with the selected slice
                min_key, max_key = ned[0]
                if selector.start is not None and selector.start > min_key:
                    min_key = selector.start
                if selector.stop is not None and selector.stop < max_key:
                    max_key = selector.stop
                selector = slice(min_key, max_key)
            dim_selector_indices[i] = selector

        if self.tensor_kind is not None:
            tensor_kind = self.tensor_kind
        elif not self.array.schema.sparse:
            tensor_kind = TensorKind.DENSE
        elif not all(
            np.issubdtype(self.array.dim(dim).dtype, np.integer) for dim in all_dims[1:]
        ):
            tensor_kind = TensorKind.RAGGED
        else:
            tensor_kind = TensorKind.SPARSE_COO

        tensor_schema_kwargs = dict(
            kind=tensor_kind,
            _array=self.array,
            _key_dim_index=key_dim_index,
            _fields=tuple(final_fields),
            _all_dims=tuple(all_dims),
            _ned=tuple(ned),
            _dim_selectors=dim_selector_indices,
            _query_kwargs={"attrs": tuple(attrs), "dims": tuple(dims)},
        )
        object.__setattr__(self, "_tensor_schema_kwargs", tensor_schema_kwargs)

    @property
    def tensor_schema(self) -> "TensorSchema[Any]":
        # local import to avoid cyclic import
        from ._tensor_schema import TensorSchemaFactories

        kwargs = self._tensor_schema_kwargs
        factory = TensorSchemaFactories[self.array.schema.sparse, kwargs["kind"]]
        return factory(**kwargs)



from functools import singledispatch
from operator import methodcaller
from typing import Any

import numpy as np
import scipy.sparse
import sparse

from .base import TensorSchema
from .mapped import MappedTensorSchema
from .sparse import SparseTensorSchema


def SparseToDenseTensorSchema(**kwargs: Any) -> TensorSchema[np.ndarray]:
    """
    Return a TensorSchema for reading sparse TileDB arrays as (dense) Numpy arrays.
    """
    return MappedTensorSchema(SparseTensorSchema(**kwargs), to_dense)


@singledispatch
def to_dense(sa: Any) -> np.ndarray:
    """Create a Numpy array from a sparse array"""
    raise NotImplementedError


to_dense.register(sparse.COO)(methodcaller("todense"))  # type: ignore
to_dense.register(scipy.sparse.csr_matrix)(methodcaller("toarray"))  # type: ignore



from __future__ import annotations

import itertools as it
from abc import ABC, abstractmethod
from bisect import bisect
from dataclasses import dataclass
from typing import Any, Generic, Iterable, Mapping, TypeVar, cast

import numpy as np

__all__ = [
    "InclusiveRange",
    "IntRange",
    "ConstrainedPartitionsIntRange",
    "WeightedRange",
]

V = TypeVar("V")
W = TypeVar("W")
VDtype = TypeVar("VDtype", bound=np.generic)
WDtype = TypeVar("WDtype", bound=np.number)


class InclusiveRange(ABC, Generic[V, W]):
    """
    Base abstract class for finite ranges that include both ends.

    An InclusiveRange `R` is a sorted set of comparable elements of type `V`. Every member
    `m` of `R` satisfies the inequality `R.min <= m <= R.max`.

    As a set, an InclusiveRange does not contain duplicates. However, range members may
    be associated with numeric weights of type `W`. This way duplicate members can be
    represented as unique members with weight equal to the number of occurrences.
    """

    __slots__: Iterable[str] = ()

    @property
    @abstractmethod
    def min(self) -> V:
        """Lower bound of this range."""

    @property
    @abstractmethod
    def max(self) -> V:
        """Upper bound of this range."""

    @property
    @abstractmethod
    def weight(self) -> W:
        """Total weight of the members of this range."""

    @property
    @abstractmethod
    def values(self) -> np.ndarray[V, Any]:
        """Unique sorted values of this range."""

    @abstractmethod
    def indices(self, values: np.ndarray[V, Any]) -> np.ndarray[np.int_, Any]:
        """Get the (0-based) indices of the given values in this range.

        :raises ValueError: If any of the given values is not in this range.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Number of unique members in this range."""

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Check if this range equals to another object.

        Two ranges are equal if they consist of the same values with the same weights.
        """

    def equal_values(self, other: InclusiveRange[V, W]) -> bool:
        """Check if two ranges consist of the same values."""
        return (
            len(self) == len(other)
            and self.min == other.min
            and self.max == other.max
            and np.all(self.values == other.values)
        )

    @abstractmethod
    def partition_by_count(self, k: int) -> Iterable[InclusiveRange[V, W]]:
        """Partition this range into `k` subranges of approximately equal weight."""

    @abstractmethod
    def partition_by_weight(self, max_weight: W) -> Iterable[InclusiveRange[V, W]]:
        """
        Partition this range into the minimum number of subranges so that each subrange
        weight is at most `max_weight`.
        """

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            slot: getattr(self, slot)
            for cls in self.__class__.mro()
            for slot in getattr(cls, "__slots__", ())
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        for slot, value in state.items():
            object.__setattr__(self, slot, value)


@dataclass(frozen=True)
class IntRange(InclusiveRange[int, int]):
    """
    An inclusive range of consecutive integers.
    """

    __slots__ = ("min", "max")
    min: int
    max: int

    def __post_init__(self) -> None:
        assert self.min <= self.max

    @property
    def weight(self) -> int:
        return len(self)

    @property
    def values(self) -> np.ndarray[int, Any]:
        return np.arange(self.min, self.max + 1)

    def indices(self, values: np.ndarray[int, Any]) -> np.ndarray[np.int_, Any]:
        indices = values - self.min
        if not (
            np.issubsctype(indices, np.integer)
            and np.all(indices >= 0)
            and np.all(indices < len(self))
        ):
            raise ValueError(f"Values not in the range {self}")
        return indices

    def __len__(self) -> int:
        return self.max - self.min + 1

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IntRange):
            return NotImplemented
        return self.min == other.min and self.max == other.max

    def equal_values(self, other: InclusiveRange[int, int]) -> bool:
        if not isinstance(other, IntRange):
            return super().equal_values(other)
        return self.min == other.min and self.max == other.max

    def partition_by_count(self, k: int) -> Iterable[IntRange]:
        n = len(self)
        if not (1 <= k <= n):
            raise ValueError(
                f"Cannot partition range of {n} members into {k} partitions"
            )
        d, m = divmod(n, k)
        # the first m partitions have length d+1 and the rest have length d
        lengths = it.chain(it.repeat(d + 1, m), it.repeat(d, k - m))
        yield from self._partition_by_lengths(lengths)

    def partition_by_weight(self, max_weight: int) -> Iterable[IntRange]:
        d, m = divmod(len(self), max_weight)
        # all partitions have length max_weight, with the possible exception of the last
        # partition that has length m (if m > 0)
        lengths = it.chain(it.repeat(max_weight, d), it.repeat(m, m > 0))
        yield from self._partition_by_lengths(lengths)

    def _partition_by_lengths(self, lengths: Iterable[int]) -> Iterable[IntRange]:
        start = self.min
        for length in lengths:
            next_start = start + length
            yield IntRange(start, next_start - 1)
            start = next_start


@dataclass(frozen=True)
class ConstrainedPartitionsIntRange(IntRange):
    """An IntRange that can be partitioned at specific start offsets."""

    __slots__ = ("start_offsets",)
    start_offsets: range

    def partition_by_count(self, k: int) -> Iterable[ConstrainedPartitionsIntRange]:
        start_offsets = self.start_offsets
        # find how many start offsets are in the (min, max] range
        i = bisect(start_offsets, self.min)
        j = bisect(start_offsets, self.max)
        num_offsets = len(start_offsets[i:j])
        if k > num_offsets + 1:  # + 1 to account for self.min as the first offset
            raise ValueError(f"Cannot partition range into {k} partitions")

        start = self.min
        for i in range(k, 1, -1):
            target_length = round((self.max - start + 1) / i)
            partition = self._next_partition(start, target_length)
            yield partition
            start = partition.max + 1
        yield ConstrainedPartitionsIntRange(start, self.max, start_offsets)

    def partition_by_weight(
        self, max_weight: int
    ) -> Iterable[ConstrainedPartitionsIntRange]:
        start_offsets = self.start_offsets
        if max_weight < start_offsets.step:
            raise ValueError(
                f"Cannot partition range with max weight={max_weight}: "
                f"start_offsets.step={start_offsets.step}"
            )

        start = self.min
        max_start = self.max - max_weight
        while start <= max_start:
            partition = self._next_partition(start, max_weight, is_max_length=True)
            yield partition
            start = partition.max + 1
        yield ConstrainedPartitionsIntRange(start, self.max, start_offsets)

    def _next_partition(
        self, start: int, target_length: int, is_max_length: bool = False
    ) -> ConstrainedPartitionsIntRange:
        start_offsets = self.start_offsets
        next_start = start + target_length
        i = bisect(start_offsets, next_start)
        assert i > 0
        # next start is start_offsets[i - 1] if the target_length is a hard upper bound
        # otherwise next_start is either start_offsets[i - 1] or start_offsets[i],
        # depending on which one is closer to the target next start
        if (
            is_max_length
            or i == len(start_offsets)
            or next_start - start_offsets[i - 1] < start_offsets[i] - next_start
        ):
            next_start = start_offsets[i - 1]
        else:
            next_start = start_offsets[i]
        return ConstrainedPartitionsIntRange(start, next_start - 1, start_offsets)


@dataclass(frozen=True)
class WeightedRange(InclusiveRange[VDtype, WDtype]):
    """An inclusive range of weighted comparable values."""

    __slots__ = ("values", "weights")
    values: np.ndarray[VDtype, Any]
    weights: np.ndarray[WDtype, Any]

    def __post_init__(self) -> None:
        assert self.values.ndim == self.weights.ndim == 1
        assert len(self.values) == len(self.weights)

    @classmethod
    def from_mapping(cls, mapping: Mapping[V, W]) -> WeightedRange[VDtype, WDtype]:
        """Create a WeightedRange from a mapping of (unique) values to weights."""
        unique_sorted, weights = zip(*sorted(mapping.items()))
        return cls(np.array(unique_sorted), np.array(weights))

    @property
    def min(self) -> VDtype:
        return cast(VDtype, self.values[0])

    @property
    def max(self) -> VDtype:
        return cast(VDtype, self.values[-1])

    @property
    def weight(self) -> WDtype:
        return cast(WDtype, np.sum(self.weights))

    def __len__(self) -> int:
        return len(self.values)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, WeightedRange):
            return NotImplemented
        return (
            len(self) == len(other)
            and np.all(self.values == other.values)
            and np.all(self.weights == other.weights)
        )

    def equal_values(self, other: InclusiveRange[VDtype, WDtype]) -> bool:
        if not isinstance(other, WeightedRange):
            return super().equal_values(other)
        return len(self) == len(other) and np.all(self.values == other.values)

    def indices(self, values: np.ndarray[VDtype, Any]) -> np.ndarray[np.int_, Any]:
        members = self.values
        indices = np.searchsorted(members, values)
        if not (np.max(indices) < len(members) and np.all(members[indices] == values)):
            raise ValueError(f"Values not in the range {self}")
        return indices

    def partition_by_count(self, k: int) -> Iterable[WeightedRange[VDtype, WDtype]]:
        n = len(self)
        if not (1 <= k <= n):
            raise ValueError(
                f"Cannot partition range of {n} members into {k} partitions"
            )

        values, weights = self.values, self.weights
        if k == n:
            # edge case: one partition per value
            for i in range(n):
                yield WeightedRange(values[i : i + 1], weights[i : i + 1])
            return

        total_weight = self.weight
        # target_weight: sum of
        # (a) weights of the partitions yielded so far and
        # (b) the average of the remaining total weight over the remaining partitions
        target_weight = total_weight / k
        acc_weights = np.cumsum(weights)
        start = 0
        for i in range(k - 1, 0, -1):
            # find the index of target_weight in the accumulated weights
            stop = int(np.searchsorted(acc_weights, target_weight))
            # check if it's after the start index; if not, move it right after start
            if stop <= start < n:
                stop = start + 1
            # otherwise the following inequality holds:
            # acc_weights[stop - 1] < target_weight <= acc_weights[stop]
            elif (
                acc_weights[stop] - target_weight
                < target_weight - acc_weights[stop - 1]
            ):
                # increment stop if acc_weights[stop] is closer to target_weight than
                # acc_weights[stop - 1]
                stop += 1

            # yield partition
            yield WeightedRange(values[start:stop], weights[start:stop])

            # update target_weight and start index
            target_weight = acc_weights[stop - 1]
            target_weight += (total_weight - target_weight) / i
            start = stop

        # yield last partition
        yield WeightedRange(values[start:], weights[start:])

    def partition_by_weight(
        self, max_weight: WDtype
    ) -> Iterable[WeightedRange[VDtype, WDtype]]:
        values, weights = self.values, self.weights
        if max_weight < np.max(weights):
            raise ValueError(
                f"Cannot partition range with max weight={max_weight}: "
                f"max value weight={np.max(weights)}"
            )
        # target_weight: sum of weights of the partitions yielded so far plus max_weight
        target_weight = max_weight
        acc_weights = np.cumsum(weights)
        start = 0
        n = len(acc_weights)
        while start < n:
            # find the index of target_weight in the accumulated weights
            stop = int(np.searchsorted(acc_weights, target_weight, side="right"))
            # yield partition
            yield WeightedRange(values[start:stop], weights[start:stop])
            # update target_weight and start index
            target_weight = acc_weights[stop - 1] + max_weight
            start = stop



import itertools as it
import os
import uuid
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Optional, Sequence

import numpy as np
import tensorflow as tf
import torch

import tiledb
from tiledb.ml.readers.types import ArrayParams, TensorKind


@dataclass(frozen=True)
class ArraySpec:
    sparse: bool
    shape: Sequence[int]
    key_dim: int
    key_dim_dtype: np.dtype
    non_key_dim_dtype: np.dtype
    num_fields: int
    tensor_kind: Optional[TensorKind]

    @staticmethod
    def combinations(
        *,
        sparse=(True, False),
        shape=((107, 10), (107, 10, 3)),
        key_dim=(0, 1),
        key_dim_dtype=(
            np.dtype(np.int32),
            np.dtype("datetime64[D]"),
            np.dtype(np.bytes_),
        ),
        non_key_dim_dtype=(np.dtype(np.int32), np.dtype(np.float32)),
        num_fields=(0, 1),
        tensor_kind=(None, TensorKind.DENSE),
    ):
        for combo in it.product(
            sparse,
            shape,
            key_dim,
            key_dim_dtype,
            non_key_dim_dtype,
            num_fields,
            tensor_kind,
        ):
            spec = ArraySpec(*combo)
            if not spec.sparse:  # dense array
                # all dimension dtypes must be integer
                if not all(
                    np.issubdtype(dtype, np.integer)
                    for dtype in (spec.key_dim_dtype, spec.non_key_dim_dtype)
                ):
                    continue
                # tensor_kind must be DENSE or None (which defaults to DENSE here)
                # don't need to check both DENSE and None, just one of them
                if spec.tensor_kind is not None:
                    continue
            else:  # sparse array
                # for non-RAGGED tensors, non_key_dim_dtype must be integer
                if spec.tensor_kind not in (None, TensorKind.RAGGED):
                    if not np.issubdtype(spec.non_key_dim_dtype, np.integer):
                        continue
            yield spec


@contextmanager
def ingest_in_tiledb(tmpdir, spec):
    """Context manager for ingesting data into TileDB."""
    array_uuid = str(uuid.uuid4())
    uri = os.path.join(tmpdir, array_uuid)
    data = original_data = _rand_array(spec.shape, spec.sparse)
    if spec.key_dim > 0:
        data = np.moveaxis(data, 0, spec.key_dim)
    data_idx = np.arange(data.size).reshape(data.shape)

    transforms = []
    for i in range(data.ndim):
        n = data.shape[i]
        dtype = spec.key_dim_dtype if i == spec.key_dim else spec.non_key_dim_dtype
        if np.issubdtype(dtype, np.number):
            # set the domain to (-n/2, n/2) to test negative domain indexing
            min_value = -(n // 2)
        elif np.issubdtype(dtype, np.datetime64):
            min_value = np.datetime64("2022-06-15")
        elif np.issubdtype(dtype, np.bytes_):
            min_value = b"a"
        else:
            assert False, dtype
        transforms.append(_IndexTransformer(f"dim_{i}", n, min_value, dtype))

    dims = [transform.dim for transform in transforms]
    attrs = [
        tiledb.Attr(name="data", dtype=np.float32),
        tiledb.Attr(name="idx", dtype=np.int16),
    ]
    schema = tiledb.ArraySchema(
        domain=tiledb.Domain(*dims), attrs=attrs, sparse=spec.sparse
    )
    tiledb.Array.create(uri, schema)

    with tiledb.open(uri, "w") as tiledb_array:
        if spec.sparse:
            nz_idxs = np.nonzero(data)
            dim_idxs = tuple(
                transform(idx) for transform, idx in zip(transforms, nz_idxs)
            )
            tiledb_array[dim_idxs] = {"data": data[nz_idxs], "idx": data_idx[nz_idxs]}
        else:
            tiledb_array[:] = {"data": data, "idx": data_idx}

    all_fields = [f.name for f in dims + attrs]
    # exclude the key dimension from the fields if it is not an integer
    if not np.issubdtype(spec.key_dim_dtype, np.integer):
        del all_fields[spec.key_dim]
    fields = np.random.choice(all_fields, size=spec.num_fields, replace=False).tolist()

    with tiledb.open(uri) as array:
        params = ArrayParams(array, spec.key_dim, fields, tensor_kind=spec.tensor_kind)
        yield params, original_data


def _rand_array(shape: Sequence[int], sparse: bool = False) -> np.ndarray:
    """Create a random array of the given shape.

    :param shape: Shape of the array.
    :param sparse:
      - If false, all values will be in the (0, 1) range.
      - If true, only one value per row will be non-zero, the rest will be 0.
    """
    if not sparse:
        return np.random.random(shape)

    rows, cols = shape[0], np.prod(shape[1:])
    a = np.zeros((rows, cols))
    col_idxs = np.random.choice(cols, size=rows)
    a[np.arange(rows), col_idxs] = np.random.random(rows)
    return a.reshape(shape)


@dataclass(frozen=True)
class _IndexTransformer:
    name: str
    size: int
    min_value: Any
    dtype: np.dtype

    @property
    def dim(self):
        return tiledb.Dim(
            name=self.name,
            domain=(self.min_value, self(self.size - 1)),
            tile=np.random.randint(1, self.size),
            dtype=self.dtype,
        )

    def __call__(self, idx):
        if isinstance(self.min_value, bytes):
            transformed_idx = _bytes_to_int(self.min_value) + idx
            if isinstance(transformed_idx, np.ndarray):
                int_to_bytes = np.vectorize(_int_to_bytes)
            else:
                int_to_bytes = _int_to_bytes
            return int_to_bytes(transformed_idx)
        else:
            return self.min_value + idx


def _bytes_to_int(data: bytes) -> int:
    s = 0
    for i, b in enumerate(reversed(data)):
        s += b * 256**i
    return s


def _int_to_bytes(n: int) -> bytes:
    s = bytearray()
    while n > 0:
        n, m = divmod(n, 256)
        s.append(m)
    s.reverse()
    return bytes(s)


def validate_tensor_generator(generator, schemas, specs, batch_size):
    if not isinstance(schemas, Sequence):
        schemas = [schemas]
    if not isinstance(specs, Sequence):
        specs = [specs]
    for batches in generator:
        if len(schemas) == 1:
            batches = [batches]
        assert len(schemas) == len(specs) == len(batches)
        for schema, spec, batch in zip(schemas, specs, batches):
            if not isinstance(batch, Sequence):
                batch = [batch]
            for tensor in batch:
                _validate_tensor(tensor, schema.kind, spec.shape[1:], batch_size)


def _validate_tensor(tensor, schema_tensor_kind, spec_row_shape, batch_size):
    tensor_kind = get_tensor_kind(tensor)
    if schema_tensor_kind is not TensorKind.RAGGED:
        assert tensor_kind is schema_tensor_kind
        if batch_size is not None:
            # batched tensor: the length of the first dimension is the size of the batch,
            # which may be smaller than the requested batch_size
            num_rows, *row_shape = tensor.shape
            assert num_rows <= batch_size, (num_rows, batch_size)
        elif schema_tensor_kind is TensorKind.SPARSE_CSR and len(spec_row_shape) == 1:
            # unbatched CSR 1D tensor has (1, N) shape
            num_rows, *row_shape = tensor.shape
            assert num_rows == 1, num_rows
        else:
            # unbatched non-CSR tensor: num_rows is implicitly 1
            row_shape = tensor.shape
        assert tuple(row_shape) == spec_row_shape
    else:
        # every ragged tensor row has at most `max_row_len` elements, the product of all
        # non-key dimension sizes
        max_row_len = np.prod(spec_row_shape)
        if batch_size is not None:
            assert tensor_kind is schema_tensor_kind
            row_lens = tuple(map(len, tensor))
            assert all(row_len <= max_row_len for row_len in row_lens)
            num_rows = len(row_lens)
            assert num_rows <= batch_size, (num_rows, batch_size)
        else:
            # in case of no batching, a 1D dense tensor is returned instead of a ragged
            # tensor with a single nested tensor
            assert tensor_kind is TensorKind.DENSE
            assert len(tensor) <= max_row_len


def assert_tensors_almost_equal_array(
    batches, array, schema_tensor_kind, batch_size, to_dense
):
    if schema_tensor_kind is TensorKind.RAGGED:
        tensor_kind = get_tensor_kind(batches[0])
        if tensor_kind is TensorKind.RAGGED:
            tensors = [tensor for batch in batches for tensor in batch]
        else:
            assert tensor_kind is TensorKind.DENSE
            tensors = batches
        # compare each tensor with the non-zero values of the respective array row
        assert len(tensors) == len(array)
        for tensor_row, row in zip(tensors, array):
            np.testing.assert_array_almost_equal(tensor_row, row[np.nonzero(row)])
    else:
        if schema_tensor_kind in (TensorKind.SPARSE_COO, TensorKind.SPARSE_CSR):
            batches = list(map(to_dense, batches))
        if batch_size is not None or schema_tensor_kind is TensorKind.SPARSE_CSR:
            np.testing.assert_array_almost_equal(np.concatenate(batches), array)
        else:
            np.testing.assert_array_almost_equal(np.stack(batches), array)


def get_tensor_kind(tensor) -> TensorKind:
    if isinstance(tensor, tf.Tensor):
        return TensorKind.DENSE
    if isinstance(tensor, tf.SparseTensor):
        return TensorKind.SPARSE_COO
    if isinstance(tensor, tf.RaggedTensor):
        return TensorKind.RAGGED
    if isinstance(tensor, torch.Tensor):
        if getattr(tensor, "is_nested", False):
            return TensorKind.RAGGED
        if tensor.layout is torch.strided:
            return TensorKind.DENSE
        if tensor.layout is torch.sparse_coo:
            return TensorKind.SPARSE_COO
        if tensor.layout is torch.sparse_csr:
            return TensorKind.SPARSE_CSR
    assert False, f"{tensor} is not a tensor"



from typing import Any, List, Mapping

import tiledb

from ..types import Selector


class KeyDimQuery:
    def __init__(
        self,
        array: tiledb.Array,
        key_dim_index: int,
        dim_selectors: Mapping[int, Selector],
        **kwargs: Any,
    ):
        self._multi_index = array.query(**kwargs).multi_index
        selectors: List[Selector] = [slice(None)] * array.ndim
        for i, selector in dim_selectors.items():
            if i == 0:
                # ignore selector for the key dimension
                continue
            elif i == key_dim_index:
                # key_dim_index got swapped with 0th index
                selectors[0] = selector
            else:
                selectors[i] = selector
        self._leading_selectors = tuple(selectors[:key_dim_index])
        self._trailing_selectors = tuple(selectors[key_dim_index + 1 :])

    def __getitem__(self, key_dim_slice: slice) -> Any:
        """Query the TileDB array by `dim_key=key_dim_slice`."""
        selectors = (*self._leading_selectors, key_dim_slice, *self._trailing_selectors)
        return self._multi_index[selectors]



"""Functionality for saving and loading PytTorch models as TileDB arrays"""
import os
import pickle
from typing import Any, Mapping, Optional

import torch
from torch.optim import Optimizer
from torch.utils.tensorboard import SummaryWriter

import tiledb

from ._base import Meta, TileDBArtifact, Timestamp


class PyTorchTileDBModel(TileDBArtifact[torch.nn.Module]):
    """
    Class that implements all functionality needed to save PyTorch models as
    TileDB arrays and load PyTorch models from TileDB arrays.
    """

    Name = "PYTORCH"
    Version = torch.__version__

    def __init__(
        self,
        uri: str,
        namespace: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        model: Optional[torch.nn.Module] = None,
        optimizer: Optional[Optimizer] = None,
    ):
        super().__init__(uri, namespace, ctx, model)
        self.optimizer = optimizer

    def save(
        self,
        *,
        meta: Optional[Meta] = None,
        summary_writer: Optional[SummaryWriter] = None,
    ) -> None:
        """
        Save a PyTorch model as a TileDB array.

        :param summary_writer:
        :param meta: Extra metadata to save in a TileDB array.
        :param summary_writer: Contains summary writer object for storing Tensorboard data.
        """
        if self.artifact is None:
            raise RuntimeError("Model is not initialized")

        # Serialize model state dictionary.
        serialized_model_dict = pickle.dumps(self.artifact.state_dict(), protocol=4)

        # Serialize model's optimizer dictionary.
        if self.optimizer:
            serialized_optimizer_dict = pickle.dumps(
                self.optimizer.state_dict(), protocol=4
            )
        else:
            serialized_optimizer_dict = b""

        # Create TileDB model array
        if not tiledb.array_exists(uri=self.uri):
            self._create_array(fields=["model", "optimizer", "tensorboard"])

        self._write_array(
            model_params={
                "model": serialized_model_dict,
                "optimizer": serialized_optimizer_dict,
            },
            tensorboard_log_dir=summary_writer.log_dir if summary_writer else None,
            meta=meta,
        )

    def load(
        self,
        *,
        model: Optional[torch.nn.Module] = None,
        optimizer: Optional[Optimizer] = None,
        timestamp: Optional[Timestamp] = None,
        callback: bool = False,
    ) -> Any:
        """
        Load switch, i.e, decide between __load (TileDB-ML<=0.8.0) or __load_v2 (TileDB-ML>0.8.0).

        :param model: A defined PyTorch model.
        :param optimizer: A defined PyTorch optimizer.
        :param timestamp: Range of timestamps to load fragments of the array which live in the specified time range.
        :param callback: Boolean variable if True will store Callback data into saved directory
        :return: A dictionary with attributes other than model or optimizer state_dict.
        """
        with tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp) as model_array:
            if self._use_legacy_schema(model_array):
                return self.__load_legacy(model_array, model, optimizer, callback)
            else:
                return self.__load(model_array, model, optimizer, callback)

    def __load_legacy(
        self,
        model_array: tiledb.Array,
        model: torch.nn.Module,
        optimizer: Optimizer,
        callback: bool,
    ) -> Optional[Mapping[str, Any]]:
        model_array_results = model_array[:]
        schema = model_array.schema

        model_state_dict = pickle.loads(model_array_results["model_state_dict"].item(0))
        optimizer_state_dict = pickle.loads(
            model_array_results["optimizer_state_dict"].item(0)
        )

        # Load model's state and optimizer dictionaries
        model.load_state_dict(model_state_dict)
        optimizer.load_state_dict(optimizer_state_dict)

        # Get the rest of the attributes
        out_dict = {}
        for idx in range(schema.nattr):
            attr_name = schema.attr(idx).name
            if (
                schema.attr(idx).name != "model_state_dict"
                and schema.attr(idx).name != "optimizer_state_dict"
            ):
                out_dict[attr_name] = pickle.loads(
                    model_array_results[attr_name].item(0)
                )

        if callback:
            try:
                with tiledb.open(f"{self.uri}-tensorboard") as tb_array:
                    for path, file_bytes in pickle.loads(
                        tb_array[:]["tensorboard_data"][0]
                    ).items():
                        log_dir = os.path.dirname(path)
                        if not os.path.exists(log_dir):
                            os.mkdir(log_dir)
                        with open(
                            os.path.join(log_dir, os.path.basename(path)), "wb"
                        ) as f:
                            f.write(file_bytes)
            except FileNotFoundError:
                print(f"Array {self.uri}-tensorboard does not exist")

        return out_dict

    def __load(
        self,
        model_array: tiledb.Array,
        model: torch.nn.Module,
        optimizer: Optimizer,
        callback: bool,
    ) -> None:
        model.load_state_dict(self._get_model_param(model_array, "model"))
        if optimizer:
            optimizer.load_state_dict(self._get_model_param(model_array, "optimizer"))
        if callback:
            self._load_tensorboard(model_array)

    def preview(self) -> str:
        """
        Create a string representation of the model.

        :return: str. A string representation of the models internal configuration.
        """
        return str(self.artifact) if self.artifact else ""



"""Tests for TileDB integration with Tensorflow Data API."""

import numpy as np
import pytest
import tensorflow as tf

from tiledb.ml.readers.tensorflow import TensorflowTileDBDataset, TensorKind

from .utils import (
    ArraySpec,
    assert_tensors_almost_equal_array,
    ingest_in_tiledb,
    validate_tensor_generator,
)


def dataset_batching_shuffling(dataset, batch_size=None, shuffle_buffer_size=0):
    if shuffle_buffer_size > 0:
        dataset = dataset.shuffle(shuffle_buffer_size)
    if batch_size is not None:
        dataset = dataset.batch(batch_size)
    return dataset


class TestTensorflowTileDBDataset:
    @pytest.mark.parametrize("spec", ArraySpec.combinations())
    @pytest.mark.parametrize("batch_size", [8, None])
    @pytest.mark.parametrize("shuffle_buffer_size", [0, 16])
    @pytest.mark.parametrize("num_workers", [0, 2])
    def test_dataset(self, tmpdir, spec, batch_size, shuffle_buffer_size, num_workers):
        def test(*all_array_params):
            dataset = dataset_batching_shuffling(
                TensorflowTileDBDataset(*all_array_params, num_workers=num_workers),
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
            )
            assert isinstance(dataset, tf.data.Dataset)
            validate_tensor_generator(
                dataset,
                [params.tensor_schema for params in all_array_params],
                [spec] * len(all_array_params),
                batch_size,
            )

        with ingest_in_tiledb(tmpdir, spec) as (x_params, x_data):
            # load tensors from single array
            test(x_params)
            with ingest_in_tiledb(tmpdir, spec) as (y_params, y_data):
                # load tensors from two arrays
                test(x_params, y_params)

    @pytest.mark.parametrize("spec", ArraySpec.combinations())
    def test_unequal_key_ranges(self, tmpdir, spec):
        y_spec = ArraySpec(
            # Add one extra key on Y
            shape=(spec.shape[0] + 1, *spec.shape[1:]),
            sparse=spec.sparse,
            key_dim=spec.key_dim,
            key_dim_dtype=spec.key_dim_dtype,
            non_key_dim_dtype=spec.non_key_dim_dtype,
            num_fields=spec.num_fields,
            tensor_kind=spec.tensor_kind,
        )
        with ingest_in_tiledb(tmpdir, spec) as (x_params, x_data):
            with ingest_in_tiledb(tmpdir, y_spec) as (y_params, y_data):
                with pytest.raises(ValueError) as ex:
                    TensorflowTileDBDataset(x_params, y_params)
                assert "All arrays must have the same key range" in str(ex.value)

    @pytest.mark.parametrize("spec", ArraySpec.combinations(num_fields=[0]))
    @pytest.mark.parametrize("batch_size", [8, None])
    def test_order(self, tmpdir, spec, batch_size):
        """Test we can read the data in the same order as written.

        The order is guaranteed only for sequential processing (num_workers=0) and
        no shuffling (shuffle_buffer_size=0).
        """
        to_dense = tf.sparse.to_dense
        with ingest_in_tiledb(tmpdir, spec) as (params, data):
            dataset = TensorflowTileDBDataset(params)
            dataset = dataset_batching_shuffling(dataset=dataset, batch_size=batch_size)
            # since num_fields is 0, fields are all the array attributes of each array
            # the first item of each batch corresponds to the first attribute (="data")
            batches = [tensors[0] for tensors in dataset]
            assert_tensors_almost_equal_array(
                batches, data, params.tensor_schema.kind, batch_size, to_dense
            )

    @pytest.mark.parametrize(
        "spec",
        ArraySpec.combinations(
            sparse=[True],
            non_key_dim_dtype=[np.dtype(np.int32)],
            tensor_kind=[TensorKind.SPARSE_CSR],
        ),
    )
    def test_csr(self, tmpdir, spec):
        with ingest_in_tiledb(tmpdir, spec) as (params, data):
            with pytest.raises(NotImplementedError) as ex:
                TensorflowTileDBDataset(params)
            assert "TensorKind.SPARSE_CSR tensors not supported" in str(ex.value)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Sequence, TypeVar

import numpy as np
import scipy.sparse
import sparse
import torch

from ._tensor_schema import TensorSchema
from .types import TensorKind

try:
    nested_tensor = torch.nested.nested_tensor
except AttributeError:
    nested_tensor = getattr(torch, "nested_tensor", None)


T = TypeVar("T")


class Collator(ABC, Generic[T]):
    """Mapper of instances of type `T` to `torch.Tensor`s"""

    @abstractmethod
    def convert(self, value: T) -> torch.Tensor:
        """Convert a tensor-like object of type `T` into a `torch.Tensor`."""

    @abstractmethod
    def collate(self, batch: Sequence[T]) -> torch.Tensor:
        """
        Collate a batch of tensor-like objects of type `T` into a `torch.Tensor`
        with an additional outer dimension - batch size.
        """

    @classmethod
    def from_schemas(cls, *schemas: TensorSchema[Any]) -> Collator[Any]:
        """
        Return an appropriate Collator for collating instances generated by the
        `iter_tensors()` method of the given schema(s).
        """
        if len(schemas) > 1:
            return RowCollator(tuple(map(cls.from_schemas, schemas)))

        schema = schemas[0]
        collator: Collator[Any]
        if schema.kind is TensorKind.DENSE:
            collator = NumpyArrayCollator()
        elif schema.kind is TensorKind.RAGGED:
            collator = NumpyArrayCollator(to_nested=True)
        else:
            assert schema.kind in (TensorKind.SPARSE_COO, TensorKind.SPARSE_CSR)
            to_csr = schema.kind is TensorKind.SPARSE_CSR
            # The sparse arrays to be converted/collated are scipy.sparse.csr_matrix if
            # (and only if) the schema ndim == 2
            if len(schema.shape) == 2:
                collator = ScipySparseCSRCollator(to_csr)
            else:
                collator = SparseCOOCollator(to_csr)

        num_fields = schema.num_fields
        return RowCollator((collator,) * num_fields) if num_fields > 1 else collator


@dataclass(frozen=True)
class RowCollator(Collator[Sequence[Any]]):
    """
    Collator of "row" tuples.

    - All rows must have the same length.
    - The i-th element of every row (i.e. the i-th "column") must have the same type.
    - The i-th column values are collated by the i-th collator given in the constructor.
    """

    column_collators: Sequence[Collator[Any]]

    def convert(self, value: Sequence[Any]) -> Sequence[torch.Tensor]:
        assert len(value) == len(self.column_collators)
        return tuple(
            collator.convert(item)
            for collator, item in zip(self.column_collators, value)
        )

    def collate(self, batch: Sequence[Sequence[Any]]) -> Sequence[torch.Tensor]:
        columns = tuple(zip(*batch))
        assert len(columns) == len(self.column_collators)
        return tuple(
            collator.collate(column)
            for collator, column in zip(self.column_collators, columns)
        )


@dataclass(frozen=True)
class NumpyArrayCollator(Collator[np.ndarray]):
    """Collator of Numpy arrays

    to_nested: If true, collate 1D Numpy arrays with possibly different length to nested
        `torch.Tensor`s. Otherwise, all the arrays to be collated must have the same shape
    """

    to_nested: bool = False

    def convert(self, value: np.ndarray) -> torch.Tensor:
        return torch.from_numpy(value)

    def collate(self, batch: Sequence[np.ndarray]) -> torch.Tensor:
        if self.to_nested:
            return nested_tensor(list(map(torch.from_numpy, batch)))
        else:
            return torch.from_numpy(np.stack(batch))


@dataclass(frozen=True)
class SparseCOOCollator(Collator[sparse.COO]):
    """Collator of sparse.COO instances

    to_csr: Collate to `torch.Tensor`s with `sparse_csr` layout if true and `sparse_coo`
        layout if False
    """

    to_csr: bool = False

    def convert(self, value: sparse.COO) -> torch.Tensor:
        if self.to_csr:
            csr = value.tocsr()
            return torch.sparse_csr_tensor(csr.indptr, csr.indices, csr.data, csr.shape)
        else:
            return torch.sparse_coo_tensor(value.coords, value.data, value.shape)

    def collate(self, batch: Sequence[sparse.COO]) -> torch.Tensor:
        return self.convert(sparse.stack(batch))


@dataclass(frozen=True)
class ScipySparseCSRCollator(Collator[scipy.sparse.csr_matrix]):
    """Collator of `scipy.sparse.csr_matrix` instances

    to_csr: Collate to `torch.Tensor`s with `sparse_csr` layout if true and `sparse_coo`
        layout if False
    """

    to_csr: bool = False

    def convert(
        self, value: scipy.sparse.csr_matrix, _collating: bool = False
    ) -> torch.Tensor:
        if self.to_csr:
            return torch.sparse_csr_tensor(
                value.indptr, value.indices, value.data, value.shape
            )

        # if not collating, convert (1, N) sparse arrays to 1D sparse COO tensors
        if not _collating and value.shape[0] == 1:
            return torch.sparse_coo_tensor(
                value.indices.reshape(1, -1), value.data, value.shape[1:]
            )

        coo = value.tocoo()
        coords = np.stack((coo.row, coo.col))
        return torch.sparse_coo_tensor(coords, coo.data, coo.shape)

    def collate(self, batch: Sequence[scipy.sparse.csr_matrix]) -> torch.Tensor:
        return self.convert(scipy.sparse.vstack(batch), _collating=True)
